"""
maestro_cap
===========

A module used to initialize a maestro experiment folder and propagate
configuration for a maestro experiment from imsi to corresponding files
of the experiment.

Maestro is a suite of tools used to create and submit tasks to
computer systems. Configuring a new maestro experiment is usually done
via command-line tools and/or the Maestro Manager program `xm`.

This imsi-maestro interface is not fully generalized (ie. there is no
"python module for maestro"). Rather, this module depends on some
initial setup in native maestro to generate the module and resource
files for an experiment. These files are then used as templates
("default"/"source"). Specific flow configuration and job resources can
be set through imsi. In other words, given a pre-made maestro experiment
with defined flow, jobs, and resources, the user can use imsi and these
functions to set specific resources or switch out one template
flow file (flow.xml) for another.

Usage notes:
- the user cannot generate a new flow/task from this module alone
(templates must be specified).
- duplicate job names in maestro are allowed (under different tasks);
for job resources configuration from imsi will be applied to all of these
jobs identically.

Maestro and its tools were written by the Canadian Meteorological Centre.

There is no planned future development for maestro, but you may still
refer to the following internal resources:

- the maestro man page (`man maestro`)
- CMC wiki https://wiki.cmc.ec.gc.ca/wiki/Maestro
- maestro git repository https://gitlab.science.gc.ca/cmoi/maestro

"""

import os
import glob
import shutil
import subprocess
import time
from pathlib import Path
import re
from xml.etree import ElementTree
import traceback

from rich.console import Console
from rich.table import Table
from rich.panel import Panel

from imsi.config_manager.config_manager import Configuration
from imsi.utils.general import parse_memory_string_to_bytes, _get_memory_unit_factors_to_bytes
from imsi.utils.nml_tools import update_env_file
from imsi.tools.time_manager.time_manager import sim_time_factory
from imsi.sequencer_interface.sequencers import Sequencer
from imsi.sequencer_interface import maestro_status


_exec_requirements = [
    'maestro',
    'expbegin',
    'makelinks',
    'nodeinfo'
]

# required by maestro
MAESTRO_FOLDERS = [
    'hub',         # host-specific dirs (static -  ~/.suites/.default_links)

    'modules',     # flow (flow/tsk/cfg) <- imsi config
    'resources',   # resources (xml) <- imsi config

    'sequencing',  # for running exp (once sequencener executed), status, output, etc.
    'listings',    # log files of maestro tasks
    'logs',        # execution logs of maestro (used by xflow UI)
    ]
# subset of folders - used when running the experiment (not used for config)
MAESTRO_FOLDERS_RUNTMP = ['sequencing', 'listings', 'logs']

class MaestroSequencerInterface(Sequencer):
    """
    The maestro sequencer interface
    """

    # define list of attributes allowed in <BATCH> tags in
    # <NODE_RESOURCE>. Note: Every attributes in the BATCH tag is
    # optional (defaults will be used by maestro if not provided).
    # See:
    #   https://wiki.cmc.ec.gc.ca/wiki/Maestro/sequencer#Batch_System_Resources
    _NODE_RESOURCE_BATCH_ATTRIBUTES = [
        'catchup', 'cpu', 'cpu_multiplier', 'immediate', 'machine',
        'memory', 'mpi', 'queue', 'soumet_args', 'wallclock'
        ]

    # imsi -> maestro for <BATCH> attributes
    _NODE_RESOURCE_BATCH_ATTRIBUTE_ALIASES = {
        'processors': 'cpu'
    }

    def __init__(self):
        # v not great v
        _check_execs_available()

    def _check_if_setup(self, seq_exp_home: str):
        # rough check to see if setup already run
        # (weak check for some paths that exist)
        modules_path = os.path.join(seq_exp_home, 'modules')
        _required_structure = [os.path.join(seq_exp_home, 'EntryModule'), modules_path]
        _empty = False
        if os.path.exists(modules_path):
            if len(os.listdir(modules_path)) == 0:
                _empty = True

        if not any([os.path.exists(f) for f in _required_structure]) or _empty:
            raise OSError("maestro sequencer folder not setup; can't continue configuration.")

    def setup(self, configuration: Configuration, force=True):
        """Setup the maestro experiment folder."""
        # Note: this setup function currently mimics most of the
        # functionality and function structure of the CanESM system
        # setup-maestro script in CCCma_tools/maestro-suite/default-config-canesm

        runid = configuration.setup_params.runid
        maestro_config = configuration.sequencing.sequencer

        exp_flow = maestro_config['baseflows']

        maestro_src = exp_flow['maestro_defaults_src']          # source of templates
        maestro_dst = maestro_config['SEQ_EXP_HOME']            # destination for setup
        modules_provided = exp_flow['flow_definitions'].keys()  # ie folder names

        setup_maestro_from_src(maestro_src, maestro_dst, modules_provided, runid=runid)

    def config(self, configuration: Configuration, **kwargs):
        """Configure maestro experiment."""
        # Note: the original config-maestro script in
        #   CCCma_tools/maestro-suite/default-config-canesm
        # does three main "steps":
        #   - sets experiment resources (to resources.def and node
        #     resource xml files, via set_resources())
        #   - sets up the experiment flow (related to flow.xml files
        #      for each module, via set_experiment_flow())
        #   - sets experiment options (to ExpOptions.xml, via
        #     set_expoptions_file())
        # The steps can be done in any order.

        maestro_config = configuration.sequencing.sequencer
        flow_config = configuration.sequencing.sequencing_flow

        exp_flow = maestro_config['baseflows']
        flow_files = exp_flow['flow_definitions']

        # TODO consider removing this imsi dependency just to access
        # the number of model/postproc iterations
        n_model_chunks, n_postproc_chunks = _get_run_chunking(
            configuration.sequencing.model_dump()['run_dates']
        )

        target_dependency_loop_ratio = n_model_chunks/n_postproc_chunks  # dynamic adj

        if not float(target_dependency_loop_ratio).is_integer():
            raise ValueError(f'(number of model chunks) / (number of postproc chunks) must be integer greater than one, not {target_dependency_loop_ratio};'+\
                              ' check settings under "run_dates".')

        #RD: I'm not sure what conditions will trigger this. If postproc_chunk_size > run length, it will still have 1 chunk
        if n_postproc_chunks < 1:
            raise ValueError(f'number of postproc chunks must be integer greater than one, not {n_postproc_chunks};'+\
                              ' check settings under "run_dates".')


        # folder structure
        # root_dir is the path to the maetro experiment, ie SEQ_EXP_HOME
        # defaults is special, but other structure under root_dir is not
        defaults_dir = maestro_config['baseflows']['maestro_defaults_src']
        exp_root_dir = maestro_config['SEQ_EXP_HOME']

        if not os.path.exists(defaults_dir):
            raise FileNotFoundError(
                f"no folder for maestro experiment template files found: {defaults_dir}"
                )

        self._check_if_setup(exp_root_dir)

        # FIXME: suppose this could be part of the experiment flow
        # definition structure, eg:
        # {"sequencing": {"sequencers": {"maestro" : {"baseflows": {
        #     "AMIP": {"canam_split_job_flow": {
        #         "flow_definitions"  : {
        #             "postproc" : "..."
        #         },
        #         "flow_adjust_loop_dependency_ratio": {
        #             "postproc" : {
        #                 "from": ["rebuild_loop", "diagnostics_loop", "wrap_up_loop"],
        #                 "to": ["model_loop", "model_loop", "model_loop"]
        #             },
        #         }
        # for now, define separate dict hardcoded here (otherwise would just use exp_flow)
        exp_flow_adj = {
            "flow_adjust_loop_dependency_ratio": {
                "postproc": {
                    "rebuild_loop": "/canesm/model/model_loop",
                    "diagnostics_loop": "/canesm/model/model_loop",
                    "wrap_up_loop": "/canesm/model/model_loop",
                }
            }
        }
        loop_adjustments = exp_flow_adj.get("flow_adjust_loop_dependency_ratio", {})

        # note: in the imsi framework, most resources for maestro jobs
        # are specific at the job-level, so this section is quite
        # small
        # FIXME still some assumptions about naming/structure here
        updates = {
            "SEQ_JOBNAME_PREFIX": generate_maestro_jobname_prefix(
                configuration.setup_params.runid
            ),
            "NUMBER_OF_POSTPROC_LOOPS": n_postproc_chunks,
            "NUMBER_OF_MODEL_LOOPS": n_model_chunks,
        }

        update_exp = maestro_config.get("experiment.cfg", {})

        # get all the names of machines that will be used to filter
        # the list available
        mset = set(
            [
                spec["resources"].get("machine", None)
                for spec in flow_config["jobs"].values()
            ]
        )
        machines = [name for name in mset if name is not None]

        platform_config = {}

        # error handling: check if information for the machine an other site
        # machines are available:
        # (this is a little messy since for the setup machine the resources
        # are set within parameters, whereas for other machines in the site they
        # are set under site)
        for m in machines:
            # Determine the source of machine configuration
            mm = configuration.machine if m == configuration.machine.name else configuration.machine.site.get(m)

            # Extract resources, handle missing cases
            m_res = getattr(mm, 'resources', None) if m == configuration.machine.name else mm.get("resources")

            if m_res is None:
                raise KeyError(f"Resources must be configured for machine '{m}'")

            platform_config[m] = {'resources': m_res}

        config_maestro_from_src(
            defaults_dir,
            exp_root_dir,
            configuration.setup_params.runid,
            flow_files,
            dependency_loop_ratio=n_model_chunks/n_postproc_chunks,
            dependency_loop_adjustments=loop_adjustments,
            dependency_loop_iter=n_postproc_chunks,
            resource_def_config=updates,
            experiment_def_config=update_exp,
            job_resource_config=flow_config["jobs"],
            platform_config=platform_config,
        )

    def submit(self, configuration: Configuration):
        """Submit a maestro job using `expbegin`.

        Note that this call will pass the current time for the datestring
        argument (`-d`), to the nearest minute only.
        """
        # Note: TODO setting the date for expbegin does NOT currently
        # mimic the behaviour of expbegin (which can automatically
        # increment the date if the same job is resubmitted, will
        # default to unix epoch time if empty, and can accept less
        # precision and pad the rest of the string with zeros).
        #
        # For now, the behaviour is to resolve the current time to the
        # nearest minute only (seconds are always "00", YYYYMMDDhhmm00)
        dtstring = get_maestro_current_datetime_string()

        expbegin_path = _get_exec_fullpath('expbegin')
        SEQ_EXP_HOME = configuration.sequencing.sequencer[
            "SEQ_EXP_HOME"
        ]

        cmd = [expbegin_path, '-e', SEQ_EXP_HOME, '-d', dtstring]

        try:
            proc = subprocess.run(cmd, capture_output=True)
            proc.check_returncode()
        except subprocess.CalledProcessError as e:
            cmd_string = ' '.join(proc.args)

            # error handling in/via expbegin is inconsistent (the 'error
            # message' might be via stdout or sterr, so capture/print both)
            out = proc.stdout.decode().strip()
            err = proc.stderr.decode().strip()
            out = out if not out else '\n' + out
            err = err if not err else '\n' + err

            raise ChildProcessError(f"Command failed: {cmd_string}\n{out}{err}") from e

    def status(self, configuration: Configuration, setup_params):
        """
        Check the status of a maestro job from the command line.

        This method retrieves and displays the status of a maestro job by reading 
        the status files from the sequencer and presenting it in a formatted 
        table with color-coded statuses.

        Args:
            configuration (Configuration): The configuration object containing 
                setup parameters and sequencing parameters.
        Notes:
            - The status is displayed in a table format using the `rich` library.
            - Only directories with names matching the pattern with 14 numeric digits are considered.
            - Only recognized statuses will be displayed in the table.
            - This command is not robustly tested for all possible maestro statuses/outputs
        """

        # recognized statuses
        color_map = {
            "begin": "green4",
            "stop": "bright_red",
            "end": "dodger_blue1",
            "catchup": "purple",
            "submit": "grey70",
        }

        root_module = maestro_status.get_root_module(Path(setup_params.work_dir))
        status_dir = Path(
            configuration.sequencing.sequencer["SEQ_EXP_HOME"],
            "sequencing",
            "status",
        )

        if not status_dir.exists():
            try:
                raise FileNotFoundError(
                    f"Status directory not found at: {status_dir}. Was the run properly submitted?"
                )
            except FileNotFoundError:
                traceback.print_exc(limit=1)
                exit(1)

        # extract experiment folder names from status directory
        date_dirs = sorted([d for d in status_dir.iterdir() if d.is_dir() and re.match(r"^\d{14}$", d.name)])[::-1]

        if len(date_dirs) == 0:
            try:
                raise FileNotFoundError(f"No experiment directories found in the status directory {status_dir}. Was the run properly submitted?")
            except FileNotFoundError:
                traceback.print_exc(limit=1)
                exit(1)

        console = Console()
        legend_str = ", ".join([f"[{v}]{k}[/{v}]" for k, v in color_map.items()])
        table = Table(show_header=True, box=None, padding=(0, 1))

        table.add_column(f"{setup_params.runid}", justify="left")
        table.add_column("Status...", justify="left")

        for date in date_dirs:
            df = maestro_status.check_entry_and_get_dataframe(root_module, date)
            # ignore empty dataframes
            if df is None:
                continue
            df = maestro_status.filter_df(df)
            # update table
            table = maestro_status.update_table(table, date, df, color_map)
        panel = Panel(table, title=f"{root_module} status [{legend_str}]", border_style="cyan")
        console.print(panel)


def _check_execs_available(names=_exec_requirements):
    # lazy check if executables are available
    for n in names:
        _get_exec_fullpath(n)

def setup_maestro_from_src(src: str, dst: str, modules: list[str], runid: str = '', force=False):
    """Setup a maestro experiment from a template source `src` at the
    destination path `dst`.

    The `src` must contain `/modules` and its contents, and a
    `EntryModule` link (optionally: `experiment.cfg`).

    Parameters:
        src : path to maestro experiment template folder. This should
            contain all required files and folders for the maestro
            experiment modules, in the expected structure.
        dst : path to destination experiment folder.
        modules : list of folder (module) names, created under `/modules`.
        runid : model run ID specified by user (usually corresponds
            to filename of $SEQ_EXP_HOME)
    """
    prepare_dst_directory(dst, force=force)
    init_folders(src, dst, modules, force=force)
    config_hub(dst, runid)

def prepare_dst_directory(dst: str, force=False):
    """Create a folder for a maestro experiment
    (the desired path for maestro's `$SEQ_EXP_HOME`).
    """
    ex = not force
    if force:
        shutil.rmtree(dst, ignore_errors=True)
    os.makedirs(dst, exist_ok=ex)

def init_folders(src: str, dst: str, modules: list[str], force=False):
    """Initialize maestro experiment folder structure at a destination
    path `dst` using a template source `src`.

    The `src` must contain `/modules` and its contents and a `EntryModule`
    link (optionally: `experiment.cfg`).

    Parameters:
        src : path to maestro experiment template folder. This should
            contain all required files and folders for the maestro
            experiment modules, in the expected structure.
        dst : path to destination experiment folder.
        modules : list of folder (module) names, created under `/modules`.
    """
    # initialize maestro folder structure
    init_maestro_experiment_folders(dst, force=force)

    copy_default_main_config(src, dst)

    # create module folders and link
    src_modules = os.path.join(src, 'modules')    # defaults
    dst_modules = os.path.join(dst, 'modules')

    entrymodule_name = get_entrymodule_name(src)
    init_maestro_module_folders(dst, entry_module=entrymodule_name, modules=modules)

    create_maestro_modules_from_src(src_modules, dst_modules)


def config_hub(seq_exp_home, experiment):
    """Create the host-specific directories in the maestro
    experiment `/hub`.

    This uses the `makelinks` program to create links to local storage
    on the appropriate machines. Note: this does not pass additional
    arguments to `makelinks`.
    """

    # setup links to machines using the makelinks maestro tool using
    # the makelinks utility (unless specified, uses
    # ~/.suites/.default_links)
    # makelinks requires:
    #   - the running location to be SEQ_EXP_HOME -> cwd
    #   - SEQ_EXP_HOME to be set as an env variable -> env
    # TODO the makelinks tool isn't overly complicated; in the future
    # it might desirable to remove this dependency in favour of
    # having the same functionality implemented here.

    makelinks_path = _get_exec_fullpath('makelinks')
    maestro_env = _set_maestro_env(seq_exp_home)

    cmd = [makelinks_path, '-f', f'--experiment={experiment}']

    try:
        proc = subprocess.run(cmd, cwd=seq_exp_home, env=maestro_env, capture_output=True)
        proc.check_returncode()
    except subprocess.CalledProcessError:
        raise Exception("Failed call: '{cmd}': {err}".format(cmd=' '.join(proc.args), err=proc.stderr))

def copy_default_main_config(src: str, dst: str):
    """Copy the configuration `experiment.cfg` file defined in
    source `src` directory to the destination `dst` directory.

    References:
        [1] https://wiki.cmc.ec.gc.ca/wiki/Maestro/experiment.cfg
    """
    exp_cfg = os.path.join(src, 'experiment.cfg')
    if os.path.exists(exp_cfg):
        shutil.copy2(exp_cfg, dst)

def init_maestro_module_folders(path: str, entry_module: str = 'module', modules: list[str] = None):
    """Initialize the experiment folder structure as required by maestro.

    Parameters:
        path: path to maestro experiment folder (top level), which contains
            the `/modules` folder (among others).
        entry_module: name of entry module, required by maestro. The
            EntryModule is then symlinked to the folder under
            `/modules/{entry_module}`
        modules: list of names of modules (subfolders). If None, will only
            include the `entry_module`.

    References:
        [1] https://wiki.cmc.ec.gc.ca/wiki/Maestro_Files_and_Folders
    """

    modules_path = os.path.join(path, 'modules')
    if os.path.exists(modules_path):
        shutil.rmtree(modules_path)
    os.makedirs(modules_path)

    if modules is None:
        modules = [entry_module]

    if entry_module not in modules:
        raise ValueError('entry_module must match one of modules provided')

    for m in modules:
        os.makedirs(os.path.join(modules_path, m))

    # this is a relative symlink
    entry_src = os.path.join('modules', entry_module)
    entry_dst = os.path.join(path, "EntryModule")
    if os.path.exists(entry_dst):
        os.remove(entry_dst)
    os.symlink(entry_src, entry_dst)

def init_maestro_experiment_folders(path: str, force=False):
    """Initialize the experiment folder structure at `path` as
    required by maestro. Only the folder structure is created;
    no folders are populated.

    References:
        [1] https://wiki.cmc.ec.gc.ca/wiki/Maestro_Files_and_Folders
    """
    for f in MAESTRO_FOLDERS:
        fpath = os.path.join(path, f)
        exists = os.path.exists(fpath)
        if f in MAESTRO_FOLDERS_RUNTMP:
            if force and exists:
                shutil.rmtree(fpath)
            os.makedirs(fpath, exist_ok=True)
        else:
            # always remake this folder
            if os.path.exists(fpath):
                shutil.rmtree(fpath)
            os.makedirs(fpath)

def get_entrymodule_name(path):
    """Given the `path` to EntryModule, get the real name of the module"""
    # note: path is the input to enforce finding the file with the
    # literal name "EntryModule" (maestro requirement)
    em = os.path.join(path, "EntryModule")
    if not os.path.exists(path) or not os.path.islink(em):
        raise FileNotFoundError(f"EntryModule not found; EntryModule must be a symlink located in {path}.")
    # get name of folder that EntryModule links to
    name = os.path.basename(os.path.realpath(em))
    return name

def create_maestro_modules_from_src(src: str, dst: str, copy=True):
    """Create maestro module structure from `src`/modules
    (default/templates) to `dst`/modules.

    If copy is True, folders and files from `src` is copied to `dst`.
    If copy is False, folder structure is copied, files (cfg, tsk) are symlinked.
    """

    if not os.path.basename(os.path.normpath(dst)) == 'modules':
        raise ValueError("dst path must be a folder named 'modules' for maestro")

    if copy:
        shutil.copytree(src, dst, dirs_exist_ok=True)
    else:
        # get list of all files (recursively)
        src_files = [p for p in glob.glob(os.path.join(src, "**"), recursive=True) if os.path.isfile(p)]

        # only include folders that are subfolders of src and immediate
        # subfolders, ie. do not include src, src/module1, src/module2, etc.
        allfolders = glob.glob(os.path.join(src, '*/**/'), recursive=True)

        # use normpath before dirname to account for potential trailing os.sep
        # TODO a pathlib.Path solution would be cleaner:
        # [f.relative_to(src) for f in Path(src).rglob(".") if not f.samefile(src) and not f.parent.samefile(src)]
        src_rel_folders = [os.path.relpath(f, src) for f in allfolders if os.path.dirname(os.path.normpath(f)) != src]

        # create local dirs within modules
        for f in src_rel_folders:
            dst_folder = os.path.join(dst, f)
            if os.path.exists(dst_folder):
                shutil.rmtree(f)
            os.makedirs(dst_folder)

        # link in version controlled files
        for f in src_files:
            s = os.path.relpath(f, src)
            os.symlink(f, os.path.join(dst, s))

def create_experiment_flow_file(src: str, dst: str):
    """Create a `flow.xml` file at the destination path `dst` by copying
    a template flow file (`xml`) `src`. While the `src` filename
    (basename) can be any name, the resulting file at `dst` is always
    `flow.xml` (maestro requirement).

    This function can be used to swap out different flow files for the
    same experiment.
    """
    if not os.path.exists(src):
        raise FileNotFoundError(f'no existing flow template file found: {src}')
    if not os.path.exists(dst):
        raise FileNotFoundError(f'no existing module set up for {dst}')
    if not os.path.basename(os.path.dirname(os.path.normpath(dst))) == 'modules':
        raise ValueError(f'incorrect destination for modules: {dst}')

    dst_flow_file = os.path.join(dst, 'flow.xml')
    shutil.copy(src, dst_flow_file)

def update_experiment_options_file(src, dst, short_name=None, display_name=None):
    """Update the attributes of the `ExpOptions` tag in the ExpOptions.xml
    file.

    The `ExpOptions.xml` is an optional file used in `xflow_overview` to
    help display experiment information.

    Note: This function has extremely limited functionality. It is
    currently only used to update two attributes on the root node
    ExpOptions tag.

    References:
        [1] https://wiki.cmc.ec.gc.ca/wiki/Maestro/ExpOptions.xml
    """

    expoptions_basename = "ExpOptions.xml"

    if short_name is None and display_name is None:
        raise ValueError('one of short_name or display_name must be specified')

    default_expoptions_file = os.path.join(src, expoptions_basename)
    if not os.path.exists(default_expoptions_file):
        raise FileNotFoundError(f'no default {default_expoptions_file} file found in {src}')

    target_expoptions_file = os.path.join(dst, expoptions_basename)
    shutil.copy2(default_expoptions_file, target_expoptions_file)

    # set attributes
    exop = ElementTree.parse(target_expoptions_file)
    root = exop.getroot()
    if short_name:
        root.set("shortName", short_name)
    if display_name:
        root.set("displayName", display_name)
    exop.write(target_expoptions_file)

def set_dependency_loop_index_by_ratio(
        flow_file: str, from_loop: str, to_loop: str, ratio: int = 1,
        n_iter : int = 1, start_iter: int = 1):
    """Set the `DEPENDS_ON` loop dependency by ratio between the index
    and the local loop, and generate each `DEPENDS_ON` element for
    the requested number of iterations.

    From the maestro docs, "a `DEPENDS_ON` element can specify
    dependencies. It tells the sequencer that the current node cannot
    be submitted until the dependency is satisfied." This function
    allows for setting iteration `N` of one task (`from_loop`) to
    depend on iteration `N` of another task (`to_loop`), ie.
    `M = r * N`, where `r` is a constant (`ratio`).

    This function edits the `flow.xml` file to 'manually' add each
    dependency loop, starting from `start_iter` index until the total
    number of iterations (`n_iter`) are reached.

    Note: This function does not check the current state of the
    `DEPENDS_ON` element. If `ratio = 1` or `n_iter = 1`, then no edits
    will be made, but this does not ensure that there is a 1-to-1
    dependency.

    Parameters:
        flow_file : path to flow file (usually `flow.xml`). Must contain
            at least one `LOOP` with name `to_loop`, which has at
            least one dependency `DEPENDS_ON` element.
        from_loop : name of loop that is the `dep_name` (dependency name)
            and is submitted for every `index`.
        to_loop : path name of loop that will be submitted for every
            `local_index`.
        ratio : ratio of loop iteration between `from_loop` and `to_loop`.
        n_iter : total number of iterations of the `to_loop` (`local_index`).
        start_iter : starting index of first iteration.

    Examples:

    In `flow.xml`, a 1-to-1 loop dependency is written as:

    ```xml
    <!-- input flow.xml -->
    <MODULE name="postproc">
        <SUBMITS sub_name="task_loop" />
        <LOOP name="task_loop">
            <DEPENDS_ON dep_name="/path/to/model_loop" index="model_loop=$((LOOP_INDEX))"
            local_index="task_loop=$((LOOP_INDEX))"
            />
        </LOOP>
    </MODULE>
    ```

    For illustration, use this function to modify the original xml
    with a 2-to-1 loop dependency (one iteration of `task_loop`
    will run after 2 iterations of `model_loop`):

    >>> set_dependency_loop_index_by_ratio('flow.xml',
    from_loop='task_loop', to_loop='/path/to/model_loop', ratio=2, n_iter=3)

    ```xml
    <!-- output flow.xml -->
    <MODULE name="postproc">
        <SUBMITS sub_name="task_loop" />
        <LOOP name="task_loop">
            <DEPENDS_ON dep_name="/path/to/model_loop" index="model_loop=2" local_index="task_loop=1" />
            <DEPENDS_ON dep_name="/path/to/model_loop" index="model_loop=4" local_index="task_loop=2" />
            <DEPENDS_ON dep_name="/path/to/model_loop" index="model_loop=6" local_index="task_loop=3" />
        </LOOP>
    </MODULE>
    ```

    References:
        [1] https://wiki.cmc.ec.gc.ca/wiki/Maestro/dependency
    """
    # Note: an issue was opened to add this feature directly into
    # maestro. The issue has since been addressed and closed, but has
    # not been tested yet by our team. Once confirmed, this function
    # may be greatly simplified if the LOOP_INDEX still needs to be set
    # dynamically.
    # https://gitlab.science.gc.ca/CMOI/maestro/-/issues/423

    #RD: this check should probably be done earlier if it is indeed necessary
    if not float(ratio).is_integer():
        raise ValueError(f'ratio must be integer greater than one, not {ratio}')

    if n_iter < 1:
        raise ValueError(f'n_iter must be integer greater than one, not {n_iter}')
    elif n_iter > 1 or ratio > 1:
        # Note: (n_iter == 1 and ratio == 1) means there's nothing to update

        tree = ElementTree.parse(flow_file)
        root = tree.getroot()

        # find the loop node
        # ensure that there is only one LOOP named loop_name
        target_loop = root.findall(f'.//LOOP[@name="{from_loop}"]')
        if not target_loop:
            raise ValueError(f'malformed flow file {flow_file}; does not contain loop {from_loop}')
        if len(target_loop) > 1:
            raise ValueError(f'malformed flow file {flow_file}; contains more than one loop for {from_loop}')
        target_loop = target_loop[0]  # safe

        # find all nodes in DEPENDS_ON node
        target_dep_nodes = list(target_loop.findall("DEPENDS_ON"))

        # make sure the dep_name for each DEPENDS_ON node matches to_loop
        filtered_nodes = []
        for node in target_dep_nodes:
            if os.path.basename(node.get('dep_name')) == os.path.basename(to_loop):
                filtered_nodes.append(node)

        if not filtered_nodes:
            raise ValueError(f'requested to_loop {to_loop} does not exist in flow file {flow_file}')

        # now that the list is filtered, use the current attribs for the
        # DEPENDS_ON nodes and reconstruct new nodes with modified attribs
        for node in filtered_nodes:

            # get attribs
            index_name = node.get('index').split('=')[0]
            local_index_name = node.get('local_index').split('=')[0]
            tail = node.tail

            # remove node
            target_loop.remove(node)

            # loop to construct new nodes based on original attribs
            for idx,i in enumerate(range(start_iter, n_iter+1)):
                target_index = int(i*ratio)
                target_index_def = f'{index_name}={target_index:d}'
                target_local_index_def = f'{local_index_name}={i:d}'

                # create xml element
                new_dep_node = ElementTree.Element('DEPENDS_ON')
                new_dep_node.set('dep_name', to_loop)
                new_dep_node.set('index', target_index_def)
                new_dep_node.set('local_index', target_local_index_def)
                new_dep_node.tail = tail

                # insert to node
                target_loop.insert(idx, new_dep_node)

        # save updated flow (overwrite)
        tree.write(flow_file)

def set_resources_def(resource_file: str, updates: dict = None):
    """Update parameters in resources.def file"""
    update_env_file(resource_file, updates=updates, key_value_only=True)

def config_node_resource_file(filename: str, resources: dict, aliases: dict = None,
                              max_bytes=None):
    """Set attributes in node resource job file.

    Parameters:
        filename : full filename of `.xml`
        resources : dict of resources to set. See Note [1].
        aliases : mapping of keys in `resources` dict to name of
            attributes allowed in maestro NODE_RESOURCES (key should
            be name in `resources`, value should be name for maestro
            node resource attribute).

    Note:
        [1] the 'memory' resource should be set by total memory (not
        memory per cpu as required by maestro; this will be handled
        internally).

    Examples:

    Given the input file `xml_file`:

    ```xml
    <NODE_RESOURCES><BATCH machine="machine" wallclock="10" /></NODE_RESOURCES>
    ```
    >>> config_node_resource_file(xml_file,
        resources={'amount_of_time': '120', 'memory': '40 g'},
        aliases={'amount_of_time': 'wallclock'})

    Output:
    ```xml
    <NODE_RESOURCES><BATCH machine="machine" wallclock="120" memory="40G" /></NODE_RESOURCES>
    ```
    """
    # note: aliases are used here just so that we can avoid renaming
    # the keys of the input resources dict

    aliases = {} if aliases is None else aliases

    # attributes for NODE_RESOURCES accepted by maestro.
    # use these as a simple check for valid config; if these are
    # variable, suggest to implement a check via `nodeinfo` instead.
    accepted_attributes = MaestroSequencerInterface._NODE_RESOURCE_BATCH_ATTRIBUTES

    try:
        contents = ElementTree.parse(filename)
    except ElementTree.ParseError as e:
        raise Exception(f'malformed resource file: {filename}') from e

    root = contents.getroot()
    if root.tag != 'NODE_RESOURCES':
        # confirm that this is the correct xml contents
        raise ValueError(f'{filename} must begin with a <NODE_RESOURCE> tag')

    batch_node = contents.findall('BATCH')
    if len(batch_node) != 1:
        raise ValueError(f'{filename} must only contain one <BATCH> tag')
    batch_node = batch_node[0]   # safe

    for k, v in resources.items():

        if k in aliases:
            k = aliases[k]

        if k not in accepted_attributes:
            # not permitted in maestro BATCH resource
            continue

        if not v:
            # empty (""), nothing to replace
            # NOTE: not used to "null" in template
            # (TODO also have not tested if empty value in xml
            # works for maestro)
            continue

        # check value of specific attributes
        if k == 'wallclock':
            _maestro_check_wallclock(v)
        elif k == 'memory':
            proc = int(resources.get('processors', 1))
            mpi = bool(int(batch_node.get('mpi', False)))
            v = _imsi_to_maestro_memory_resource(v, processors=proc, unit='G', mpi=mpi, max_bytes=max_bytes)

        batch_node.set(k, v)

    contents.write(filename)

def copy_resources_from_src(src, dst):
    # copy in the structure/files
    if not os.path.exists(dst) or not any(os.scandir(dst)):
        # if doesn't exist or empty
        # dirs_exist_ok will then only ignore parent dir (desired)
        shutil.copytree(src, dst, copy_function=shutil.copy2, dirs_exist_ok=True)
    else:
        # TODO this is the previous behaviour- should this do something different? (raise / recopy?)
        print('resource directory already populated')

def _get_run_chunking(run_dates_dict):
    # run_dates_dict is the dict required to create a SimulationTime
    # object using the from_kwargs method.
    # this could be more explicit (ie only use args needed to generate
    # chunk sizes), but using built in object for now:
    # sim_timers = SimulationTime.from_kwargs(**run_dates_dict)
    chunks = {}
    for timer_name in ['model_submission_job', 'postproc_submission_job']:
        sim_timer = sim_time_factory(run_dates_dict, timer_name)
        chunks[f"n_{sim_timer.chunk_prefix}s"] = sim_timer.NumChunks

    return chunks['n_model_chunks'], chunks['n_postproc_chunks']


def _create_basename_lookup(filelist):
    """Given a list of files (`filelist`), create lookup dictionary
    by filename (keys) to all file paths with the same filename
    (values). Each list is the full path including the original
    basename (with extension). Real/abs paths are not resolved,
    duplicates are preserved.

    Examples:
    >>> _create_basename_lookup(['/path/to/file_a', '/path/to/file_b', 'path/to/another/file_a'])
    {'file_a': ['/path/to/file_a', 'path/to/another/file_a'],
     'file_b': ['/path/to/file_b']}
    >>> _create_basename_lookup(['/path/to/file_a', '/path/to/file_a'])  # duplicates
    {'file_a': ['/path/to/file_a', '/path/to/file_a']}
    """
    basename_lookup = {}
    for f in filelist:
        fname = os.path.splitext(os.path.basename(f))[0]
        if fname in basename_lookup:
            basename_lookup[fname].append(f)
        else:
            basename_lookup[fname] = [f]
    return basename_lookup

def update_experiment_config(path: str, updates: dict = None):
    # append exerpiment variables to end of experiment.cfg
    cfg = os.path.join(path, 'experiment.cfg')
    if not os.path.exists(cfg):
        with open(cfg, 'a') as f:
            pass
    update_env_file(cfg, updates=updates, key_value_only=False)

def generate_maestro_jobname_prefix(s, sep='-') -> str:
    """Generate a jobname prefix start with `s` and ending with `sep`.
    By convention, `s` is usually the "runid" of the job. `sep` is only
    added if `s` does not already end with `sep`.
    """
    if s.endswith(sep):
        return s
    else:
        return f'{s}{sep}'

def config_maestro_from_src(
        src: str, seq_exp_home: str, runid: str, flow_files_lookup: dict,
        dependency_loop_ratio: int = 1,
        dependency_loop_iter: int = 1,
        dependency_loop_adjustments: dict = None,
        resource_def_config: dict = None,
        experiment_def_config : dict = None,
        job_resource_config: dict = None,
        platform_config: dict = None):
    """Configure the maestro experiment resources from templates (source).
    Set the main `flow.xml` files and configure the job node resource
    `.xml` files with requested resources.

    Parameters:
        src : path to maestro experiment template folder that contains
            the template files under a `/resources` folder.
        seq_exp_home : destination for maestro sequencer setup.
        runid : model run ID specified by user (usually corresponds
            to filename of $SEQ_EXP_HOME)
        flow_files_lookup : dict where keys are the name of the maestro
            experiment module and values are paths to the template flow
            files. These template flow files may have any name and will
            be renamed to `flow.xml` in the appropriate destination
            module folder when copied.
        dependency_loop_ratio : int
        dependency_loop_iter : int
        dependency_loop_adjustments : dict
        resource_def_config : where keys are names of a maestro resource
            resource in `resources.def` and values are the desired setting.
        experiment_def_config : where keys are names of a maestro setting
            to update in the experiment.cfg file.
        job_resource_config : nested dict of settings for each maestro
            job, which contains a dict of 'resources' for maestro. eg:
            {'job_name_A' : {'resources': {'wallclock': '120'}}}
    """
    validate_maestro = True
    max_bytes_fallback = 1.87E11   # bytes (= 187000 MB)

    # src / templates
    resources_src_dir = os.path.join(src, 'resources')

    # dst / experiment folder
    modules_dst_dir = os.path.join(seq_exp_home, 'modules')
    resources_dst_dir = os.path.join(seq_exp_home, 'resources')

    dlr = dependency_loop_ratio
    dli = dependency_loop_iter
    dla = dependency_loop_adjustments

    #-- set experiment options (ExpOptions)
    update_experiment_options_file(
        src, seq_exp_home,
        short_name=runid, display_name=runid
        )

    #-- set module flows (flow.xml)
    for module, src_flow_file in flow_files_lookup.items():

        create_experiment_flow_file(src_flow_file, os.path.join(modules_dst_dir, module))

        # special edits as defined in dla for dependency loops
        # (maestro DEPENDS_ON)
        if module in dla:
            # ratio of n loops of different experiment components
            # (see info in set_dependency_loop_index_by_ratio)

            if dlr > 1:
                adjust_flow = os.path.join(modules_dst_dir, module, 'flow.xml')
                for from_loop,to_loop in dla[module].items():
                    set_dependency_loop_index_by_ratio(
                        from_loop=from_loop,
                        to_loop=to_loop,
                        flow_file=adjust_flow,
                        ratio=dlr,
                        n_iter=dli
                        )


    #-- add common experiment variables
    if experiment_def_config:
        update_experiment_config(seq_exp_home, experiment_def_config)

    #-- set common task resources (resources.def)
    copy_resources_from_src(resources_src_dir, resources_dst_dir)

    # made by running copy_resources_from_src (risky?)
    resource_file = os.path.join(resources_dst_dir, 'resources.def')

    set_resources_def(resource_file=resource_file, updates=resource_def_config)

    #-- set specific job node resources (xml NODE_RESOURCES)

    # get all the xml files - use the ones in the dst experiment folder
    # (they have been copied in and now need "updating" with specific
    # configuration)
    taskfiles = glob.glob(os.path.join(resources_dst_dir, '**/*.xml'), recursive=True)

    # the lookup below handles duplicate filenames between different
    # folders. This assumes that the file contents and resources
    # requested are IDENTICAL.
    taskfile_lookup = _create_basename_lookup(taskfiles)

    # configure resources for each job
    #
    # this loop goes over the the jobs defined in the json; this means
    # there can be templates for other jobs/resources that are not
    # explicitly configured through the imsi interface.
    # note: therefore, with this loop over jobs from json, there is no need
    # to explicitly handle (skip) 'container.xml' files
    for name,spec in job_resource_config.items():

        # find corresponding task file
        try:
            taskfiles = taskfile_lookup[name]
        except KeyError:
            # if the job is defined in imsi, then it must exist in
            # maestro defaults -> raise if not
            raise FileNotFoundError(
                    "no corresponding maestro task file found "
                    f"for '{name}' in sequencing_flow; settings "
                    "will not be applied"
                    )

        resources = spec['resources']

        for tf in taskfiles:

            # if the machine is specified in the resources, then
            # it will be used for configuration of the job via
            # config_node_resource_file(). This means we can
            # grab the corresponding max_bytes memory for that machine
            # here.
            # if machine is not set in resources, then it could be
            # defined in the xml resource file itself (or not). this
            # information is only available once the xml file is read
            # (which currently happens inside the next function call),
            # which then makes it clumsy to try to determine the correct
            # max_byte memory from inside.
            # FIXME: instead, set the max_mem in this case to a reasonable
            # guess for now.
            machine_name = resources.get('machine', None)
            if machine_name is None or platform_config is None:
                max_bytes = max_bytes_fallback
            else:
                max_mem = platform_config.get(machine_name, {}).get('resources', {}).get('max_mem_per_node', None)
                if max_mem is None:
                    max_bytes = max_bytes_fallback
                else:
                    max_bytes = parse_memory_string_to_bytes(max_mem, base=10)

            config_node_resource_file(
                tf, resources=resources,
                aliases=MaestroSequencerInterface._NODE_RESOURCE_BATCH_ATTRIBUTE_ALIASES,
                max_bytes=max_bytes
            )

    if validate_maestro:
        validate_maestro_exp(seq_exp_home)

def validate_maestro_exp(seq_exp_home: str):
    """Validate maestro experiment setup using the nodeinfo tool.

    nodeinfo will catch syntax errors, unset global resources, etc.
    """

    resource_dir = os.path.join(seq_exp_home, 'resources')

    # The maestro_cap behaves such that *any* cfg/tsk/xml files in the
    # template folder are copied into the current experiment folder,
    # even if they aren't used in the specified flow.xml. These are
    # generally benign, but will throw non-zero exit status from nodeinfo.
    # The stderr from checking these nodes will start with:
    #     'Unable to get to the specified node'.
    # This can also be triggered for an arbitrary node that that
    # doesn't exist (`nodeinfo -n path/made/up`), but this won't
    # occur in this validation because only the nodes (file structure)
    # that exists in the exp folder is checked.
    err_startswith = 'Unable to get to the specified node'
    nodeinfo_path = _get_exec_fullpath('nodeinfo')
    maestro_env = _set_maestro_env(seq_exp_home)

    allfiles = glob.glob(os.path.join(resource_dir, "**/*"), recursive=True)
    exp_nodes = [os.path.splitext(os.path.relpath(f, start=resource_dir))[0] for f in allfiles if not os.path.basename(f).endswith('def') and not os.path.basename(f) == 'container.xml']

    # there isn't clear/nice way to execute multiple shell commands
    # (avoid shell=True or writing a temporary script), so loop over
    # instead (might be inefficient; look into tempfile)
    for node in exp_nodes:
        cmd = [nodeinfo_path, "-n", node]
        try:
            proc = subprocess.run(cmd, cwd=seq_exp_home, env=maestro_env, capture_output=True)
            proc.check_returncode()
        except subprocess.CalledProcessError:
            stdout = proc.stdout.decode().strip()
            stderr = proc.stderr.decode().strip()
            if not stderr.startswith(err_startswith):
                raise ChildProcessError(
                    "Maestro validation error:\nSEQ_EXP_HOME={exp}\nCalled: {cmd}\n{stdout}\n{stderr}".format(
                    exp=seq_exp_home, cmd=' '.join(proc.args),
                    stdout=stdout, stderr=stderr
                    )
                )

def _set_maestro_env(seq_exp_home: str):
    # set environment variables required by maestro - SEQ_EXP_HOME
    maestro_env = os.environ.copy()
    maestro_env['SEQ_EXP_HOME'] = seq_exp_home
    return maestro_env

def _maestro_check_wallclock(duration):
    """Check wallclock value format"""
    if not float(duration).is_integer():
        raise ValueError(f'invalid wallclock value: {duration}')

def _maestro_format_resource_memory(nbytes, unit='G', precision=0, base=2):
    """Take number of bytes `nbytes` and return a string formatted with
    the memory in the units of `unit`. The string can be used for
    configuring maestro resources (xmls).
    """
    unit_map = _get_memory_unit_factors_to_bytes(base)
    cm = nbytes / unit_map[unit]
    return '{cm:.{n}f}{unit}'.format(cm=cm, unit=unit, n=precision)

def _imsi_to_maestro_memory_resource(imsi_memory: str, processors=1, unit='G', precision=0, mpi=False, max_bytes=None):
    """Convert memory specified in imsi (total memory) to maestro
    requirement (memory per cpu/processor).

    Examples:
    >>> _imsi_to_maestro_memory_resource("120GB")
    '120G'
    >>> _imsi_to_maestro_memory_resource("480 GB", processors=240)
    '2G'

    References:
        [1] https://wiki.cmc.ec.gc.ca/wiki/Maestro/sequencer#Batch_System_Resources
        [2] https://portal.science.gc.ca/confluence/display/SCIDOCS/ord_soumet+Usage
    """
    base = 10    # maestro/ord_soumet uses base 10
    nbytes = parse_memory_string_to_bytes(imsi_memory, base=base)

    # FIXME this is not quite right but gets the behaviour
    # we want for now
    if mpi:
        cm_bytes = nbytes / processors    # mem per cpu
    else:
        cm_bytes = nbytes

    if max_bytes is not None:
        # check if cm_bytes exceeds maximum allowed
        # when max is exceeded, maestro will throw an abort in the submission
        # listing that says the memory "maximum possible is 187000MB".
        # note: may need to refactor this section once behaviour of ord_soumet
        # is better understood re: mpi
        if cm_bytes > max_bytes:
            # ord_soumet is base 10
            cm_bytes_fmt = _maestro_format_resource_memory(
                cm_bytes, unit=unit, precision=precision, base=base
                )
            max_bytes_fmt = _maestro_format_resource_memory(
                max_bytes, unit=unit, precision=precision, base=base
                )
            raise ValueError(
                f"memory per cpu ({cm_bytes_fmt}) exceeds maximum allowed ({max_bytes_fmt}); "
                f"check total memory ({imsi_memory}) and processors requested ({processors}). "
                f"(mpi = {mpi})"
                )

    fmt = _maestro_format_resource_memory(cm_bytes, unit=unit, precision=precision, base=base)

    return fmt

def _get_exec_fullpath(program_name: str) -> str:
    # return full path to a maestro executable -> `which`
    pgm = shutil.which(program_name)
    if pgm is None:
        raise FileNotFoundError(
            f"No program '{program_name}' found. For using the maestro "
            "sequencer, maestro-related tools must be available "
            "on your system."
            )
    return pgm

def get_os_time() -> time.struct_time:
    return time.localtime(time.time())

def get_maestro_current_datetime_string() -> str:
    """Generate a string of datetime information for the current date
    and time that can be used by maestro's `expbegin` tool, to the
    nearest minute (`YYYYMMDDhhmm00`).

    Note: while `expbegin` accepts a datetime string up to 14 characters (ie.
    `YYYYMMDDhhmmss`), this function only generates a datetime string
    to the nearest minute (forced to zero).
    """
    t = get_os_time()
    fmt = "%Y%m%d%H%M00"
    dts = time.strftime(fmt, t)
    return dts



from pydantic import BaseModel, model_validator, field_validator
from pathlib import Path
from imsi.utils.general import is_root_of

class SetupParams(BaseModel):
    """SetupParams configuration dataclass"""
    runid: str
    model_name: str
    experiment_name: str
    machine_name: str | None
    compiler_name: str | None
    sequencer_name: str | None
    flow_name: str | None
    postproc_profile: str | None
    work_dir: str | Path          # dev: these should really be string
    run_config_path: str | Path   #   because they are rendered as such
    source_repo: str | Path
    fetch_method: str
    source_version: str | None    # None when fetch_method !~ git
    source_path: str | Path
    imsi_config_path: str | Path
    imsi_venv: str

    # TODO construction/validation to be refined/expanded

    @model_validator(mode='after')
    def check_setup_structure(self) -> str:
        for path in [self.run_config_path, self.source_path, self.imsi_config_path]:
            isroot = is_root_of(self.work_dir, path, greedy=False)
            if not isroot:
                raise ValueError(
                    f'setup path {Path(path).name} must be within {self.work_dir}, not at {path}'
                    )
        return self



from imsi.cli.entry import cli
import shutil
# import pytest
from pathlib import Path
from click.testing import CliRunner
import os
import subprocess
from imsi.tools.time_manager.time_manager import sim_time_factory
from imsi.shell_interface.shell_timing_vars import generate_timers_config
from imsi.utils.general import write_shell_script

#WIP: testing ground for time-manager and chunk-manager routines
#Need to implement testing protocol (pytest?)
#and test a number of configurations.
#Also, diff output files and cleanup afterwards.

# Outputs a bunch of timer lists nested in the timer_dates directory
# Demonstrates some of the capability of the current time manager

output_path = Path('timer_dates')
output_path.mkdir(exist_ok=True)

config_list = [
                {"name": "12MS-12MS-Jan1", "run_segment_start_time": "6000-01-01T00:00:00",  "model_chunk_size": "P12MS", "model_internal_chunk_size": "P12MS"},
                {"name": "1MS-1MS-Jan1", "run_segment_start_time": "6000-01-01T00:00:00", "model_chunk_size": "P1MS", "model_internal_chunk_size": "P1MS"},
                {"name": "12M-12M-Jan15", "run_segment_start_time": "6000-01-15T00:00:00", "model_chunk_size": "P12M", "model_internal_chunk_size": "P12M"},
                {"name": "1M-1M-Jan15", "run_segment_start_time": "6000-01-15T00:00:00", "model_chunk_size": "P1M", "model_internal_chunk_size": "P1M"},
                {"name": "12MS-12MS-Jan15", "run_segment_start_time": "6000-01-15T00:00:00",  "model_chunk_size": "P12MS", "model_internal_chunk_size": "P12MS"},
                {"name": "1MS-1MS-Jan15", "run_segment_start_time": "6000-01-15T00:00:00", "model_chunk_size": "P1MS", "model_internal_chunk_size": "P1MS"},
                {"name": "12MS-1MS-Jan1", "run_segment_start_time": "6000-01-01T00:00:00",  "model_chunk_size": "P12MS", "model_internal_chunk_size": "P1MS"},
                {"name": "12MS-1DS-Jan1", "run_segment_start_time": "6000-01-01T00:00:00",  "model_chunk_size": "P12MS", "model_internal_chunk_size": "P1DS"},
                {"name": "1DS-1DS-Jan1-12pm", "run_segment_start_time": "6000-01-01T12:00:00",  "model_chunk_size": "P1DS", "model_internal_chunk_size": "P1DS"},
                {"name": "1D-1D-Jan1-12pm", "run_segment_start_time": "6000-01-01T12:00:00",  "model_chunk_size": "P1D", "model_internal_chunk_size": "P1D"},
              ]

# Test settings
run_dates = {
        "run_start_time": "6000-01-01T00:00:00",
        "run_stop_time": "6010-12-31T23:59:59",
        "run_segment_start_time": "6000-01-01T00:00:00",
        "run_segment_stop_time": "6010-12-31T23:59:59",
        "model_chunk_size": "P1MS",
        "model_internal_chunk_size": "P12MS",
        "postproc_chunk_size": "P12MS"
}

for config in config_list:

    run_dates.update(config)

    #First output the dates file using the time manager.
    #This file is still in use by ISS/SSS though will
    #not be used anymore by Maestro. We need to do some
    #more careful work to see if we can remove the need
    #for dates files in ISS/SSS
    run_config_path = output_path / config["name"]
    run_config_path.mkdir(exist_ok=True)

    dates_list = run_config_path / 'model_submission_job_start-stop_dates'
    if dates_list.exists():
        dates_list.unlink()
    write_shell_script(os.path.join(str(run_config_path), 'model_submission_job_start-stop_dates'),
                    generate_timers_config(run_dates, 'model_submission_job'))

    dates_list = run_config_path / 'model_inner_loop_start-stop_dates'
    if dates_list.exists():
        dates_list.unlink()
    write_shell_script(os.path.join(str(run_config_path), 'model_inner_loop_start-stop_dates'),
                    generate_timers_config(run_dates, 'model_inner_loop'))





"""This module provides classes and utilities for managing configurations,
including models, experiments, machines, compilers, and their interactions.

The module includes classes for defining configuration elements such as
Model, Experiment, Machine, and Compiler. Additionally, it provides a
Configuration class for composing these elements into a complete configuration.

ConfigManager is a class used to establish configuration objects and
facilitate saving and loading configurations for later use. The module also
includes abstract classes such as ConfigDatabase for defining an abstract
interface for a configuration database and its concrete implementation JsonConfigDatabase.

   Lastly, the module contains Factory classes (eg. CompilerFactory),
   which provide methods for creating instances of analogous classes (eg. Compiler).

The module is designed to support flexibility in configuration management,
allowing configurations to be composed, serialized, and deserialized.

This module is a WIP. It is next required to create infrastructure
modules that use the configurations created here.

Neil Swart, April 2024
"""
# TODO:
#   - Add a setup / user config (selections incoming from setup / subsequent "config" calls)
#   - Perhaps represent/process shell_parameters more explicitly

from typing import Dict
import copy
import pickle
from pydantic import BaseModel
import yaml

from imsi.utils.dict_tools import (
    parse_config_inheritance,
    update,
    recursive_lookup,
    replace_curlies_in_dict,
)
from imsi.config_manager.databases import ConfigDatabase, JsonConfigDatabase, YAMLConfigDatabase
from imsi.config_manager.schema.machine import Machine
from imsi.config_manager.schema.model import Model
from imsi.config_manager.schema.experiment import Experiment
from imsi.config_manager.schema.components import Components
from imsi.config_manager.schema.compiler import Compiler
from imsi.config_manager.schema.sequencing import Sequencing
from imsi.config_manager.schema.post_processing import PostProcessing
from imsi.config_manager.schema.setup_params import SetupParams
from imsi.config_manager.schema.utilities import Utilities


def database_factory(
    imsi_config_path: str, config_type: str = "yaml"
) -> ConfigDatabase:
    """Factory function to create a ConfigDatabase instance based on the config_type"""

    config_types = {
        "json": JsonConfigDatabase(imsi_config_path),
        "yaml": YAMLConfigDatabase(imsi_config_path),
    }

    try:
        return config_types[config_type]
    except KeyError:
        raise ValueError(f"Invalid config_type: {config_type}")


class CompilerFactory:
    """Class containing methods to instantiate an instance of Compiler.
    Like for machines, several parsing functions are required and encapsulated here.
    """

    @staticmethod
    def create_from_database(
        db: ConfigDatabase, machine: Machine, compiler_name: str = None
    ):
        compiler_name = compiler_name or machine.default_compiler
        machine.validate_compiler(compiler_name=compiler_name)
        compiler_config = db.get_parsed_config("compilers", compiler_name)
        # name is in the class, but for consistency of others return it separately
        return compiler_name, Compiler(name=compiler_name, **compiler_config)


class SequencingFactory:
    """Class containing methods to instantiate an instance of Sequencing.
    Like for machines, several parsing functions are required and encapsulated here.
    """

    @staticmethod
    def create_from_database(
        db: ConfigDatabase,
        machine: Machine,
        experiment: Experiment,
        model: Model,
        sequencer_name: str,
        flow_name: str = None,
        model_type: str = None,
    ):
        """Create a sequencing instance given the config db, the machine, sequencer and flow names"""

        sequencing_config = {}  # To build up here
        sequencing_configs = db.get_config("sequencing")  # Everything from the DB
        # Validity checks
        SequencingFactory.verify_sequencing_structure(sequencing_configs)

        # Run dates
        sequencing_config["run_dates"] = sequencing_configs["run_dates"]

        # This is the sequencer specific content (encapsulate in a function)
        sequencing_config["sequencer"] = SequencingFactory.set_sequencer_config(
            sequencer_name=sequencer_name,
            sequencers=sequencing_configs["sequencers"],
            machine_name=machine.name,
        )

        # Set default flow if it is not defined
        flow_name = flow_name or SequencingFactory.determine_default_flow(
            model_type,
            sequencer_name,
            sequencing_config["sequencer"].get("baseflows"),
            machine_specific_flows=sequencing_configs["sequencing_flow"],
            machine=machine,
            experiment=experiment,
            model=model
        )

        sequencing_config["sequencing_flow"] = SequencingFactory.set_flow_config(
            flow_name=flow_name,
            flows=sequencing_configs["sequencing_flow"],
            sequencer_name=sequencer_name,
            sequencer_config=sequencing_config["sequencer"],
            machine=machine,
            model_type=model_type,
        )

        # Resolve sequencer config. It is a bit messy at there is some iterative dependency resolving between sequencer and flow
        #  flow_name.split('-')[0] is extracting the baseflow using the name convention for flows.
        try:
            sequencing_config["sequencer"]["baseflows"] = sequencing_config[
                "sequencer"
            ]["baseflows"][model_type][flow_name.split("-")[0]]
        except:
            raise KeyError(
                "Error is setting sequencer_config, using the flow name: {flow_name}. Is the flow_name valid?"
            )
        return (sequencer_name, flow_name), Sequencing(**sequencing_config)

    @staticmethod
    def verify_sequencing_structure(sequencing_config: dict):
        """
        Checks what we got from the db includes mandatory sections
        (replaceable by schema validation?)
        """
        required_keys = ["run_dates", "sequencing_flow", "sequencers"]
        for key in required_keys:
            if key not in sequencing_config.keys():
                raise KeyError(
                    f"The key {key} is not in the 'sequencing' configuration "
                    "provided, but is a required element"
                )

    @staticmethod
    def determine_default_flow(
        model_type: str,
        sequencer_name: str,
        sequencer_baseflows: dict,
        machine_specific_flows,
        machine: Machine,
        experiment: Experiment,
        model: Model,
    ):
        """Get the default sequencing flow, which handles the selected model_type, and
        also has configuration support for the selected sequencer and machine.

        Parameters:
         model_type (str): The model configuration, e.g. ESM, AMIP or OMIP
        """
        if model_type not in sequencer_baseflows:
            supported_model_type = ", ".join(sequencer_baseflows.keys())
            raise KeyError(
                f"The selected model configuration: {model_type} is not supported by "
                f"available workflows of the selected sequencer: {sequencer_name}. "
                f"Supported model configurations are: {supported_model_type}"
            )

        # All baseflows for this sequencer, and this model_type
        base_flows = sequencer_baseflows[model_type].keys()

        # The platform/machine specific implementation of baseflows is denoted with a suffix on baseflow
        # This will return the first baseflow/machine specific flow in the list (which is constrain to be
        # only those suppoting the specific model_type)
        # Prefer explicit 'experiment' field, fall back to 'model'; error if neither present
        experiment_or_model_flow = experiment.flow or model.flow

        for baseflow in base_flows:
            if experiment_or_model_flow is not None:
                machine_specific_flow = f"{baseflow}-{experiment_or_model_flow}-{machine.default_sequencing_suffix}"
            else:
                machine_specific_flow = f"{baseflow}-{machine.default_sequencing_suffix}"
            if machine_specific_flow in machine_specific_flows:
                return machine_specific_flow

        supported_base_flows = ", ".join(base_flows)
        raise KeyError(
            f"Could not determine a default sequencing flow for sequencer '{sequencer_name}' "
            f"on machine '{machine.name}' (suffix '{machine.default_sequencing_suffix}'), "
            f"for model_type '{model_type}' and experiment or model flow '{experiment_or_model_flow}'. "
            f"Supported baseflows for this model_type are: {supported_base_flows}. "
            "Machine-specific versions for this combination are not implemented but required."
        )

    @staticmethod
    def set_sequencer_config(sequencer_name: str, sequencers: dict, machine_name: str):
        # This is the sequencer specific content (encapsulate in a function)
        if sequencer_name not in sequencers:
            supported_sequencers = ", ".join(sequencers.keys())
            raise KeyError(
                f"Selected sequencer {sequencer_name} not in list of configured sequencers {supported_sequencers}"
            )

        sequencer_config = parse_config_inheritance(sequencers, sequencer_name)

        # Validation
        if "supported_machines" not in sequencer_config:
            raise KeyError(
                f"No 'supported_machines' field in sequencer definition for {sequencer_name}"
            )
        if machine_name not in sequencer_config.get("supported_machines"):
            supported_machines = ", ".join(sequencer_config.get("supported_machines"))
            raise KeyError(
                f"Machine {machine_name} not listed as supported by sequencer {sequencer_name}. "
                f"Supported machines for {sequencer_name} are {supported_machines}. "
                "Either change machine or sequencer."
            )
        if not sequencer_config or "baseflows" not in sequencer_config:
            raise ValueError(
                f"The sequencer config for {sequencer_name} does not contain a 'baseflows' definition"
            )

        return sequencer_config

    @staticmethod
    def set_flow_config(
        flow_name: str,
        flows: dict,
        sequencer_name: str,
        sequencer_config: dict,
        machine: Machine,
        model_type: str,
    ):
        if flow_name not in flows:
            supported_flows = ", ".join(flows.keys())
            raise KeyError(
                f"Selected sequencing flow {flow_name} not in list of configured flows:"
                f"{supported_flows}"
            )
        flow_config = parse_config_inheritance(flows, flow_name)

        # Validation
        if not flow_config and "base_flow" in flow_config:
            raise ValueError(
                "The flow configuration for flow {flow_name} is not valid "
                "or does not contain the required 'base_flow'"
            )
        base_flow = flow_config.get("base_flow")

        if base_flow not in sequencer_config["baseflows"][model_type]:
            supported_base_flows = ", ".join(
                sequencer_config["baseflows"][model_type].keys()
            )
            raise ValueError(
                f"The sequencer config for {sequencer_name} does not contain a "
                f"flow definition for flow: {flow_name}. Supported (base) flows are: {supported_base_flows}"
            )
        return flow_config


class PostprocFactory:
    """Class containing methods to instantiate an instance of Postprocessing.
    Fetch default from Experiment object (preferred) or Model object (backup)
    if postproc is not set by the user.
    """

    @staticmethod
    def create_from_database(
        db: ConfigDatabase,
        model: Model,
        experiment: Experiment,
        postproc_name: str = None,
    ):
        if (
            postproc_name is None or postproc_name == ""
        ):
            postproc_name = PostprocFactory.get_default_postproc(model, experiment)

        postproc_config = db.get_parsed_config("post-processing", postproc_name)

        return postproc_name, PostProcessing(**postproc_config)

    @staticmethod
    def get_default_postproc(model, experiment) -> str:
        """Retrieve the default postproc_profile"""
        if "postproc_profile" in experiment.model_dump():
            return experiment.postproc_profile
        elif "postproc_profile" in model.model_dump():
            return model.postproc_profile
        else:
            raise ValueError(
                f"No default postproc_profile defined for {experiment.name} or {model.name}"
            )


class ExperimentFactory:
    """Class containing methods to instantiate an instance of Experiment.
    On particular check that the model and experiment are consistent
    """

    @staticmethod
    def create_from_database(db: ConfigDatabase, experiment_name: str, model_name: str):
        """Create a sequencing instance given the config db, the machine, sequencer and flow names"""

        # An issue with this check here is that if a users changes the model and does
        # imsi config, this will not be triggered, but it could be invalid (should only be done
        # via imsi set)

        experiment_data = db.get_parsed_config("experiments", experiment_name)
        experiment = Experiment(name=experiment_name, **experiment_data)
        experiment.validate_model_name(model_name)
        return experiment


class Configuration(BaseModel):
    """Container class that combines sub-configurations and serves as the goto reference
    defining the configuration of a simulation.

    """

    # maybe better to include model (header), experiment (header) and a new component object (merged)??
    model: Model
    experiment: Experiment
    components: Components
    machine: Machine
    compiler: Compiler
    postproc: PostProcessing
    setup_params: SetupParams
    utilities: Utilities
    sequencing: Sequencing
    # Add "output templates" of which "shell_params" could be one?
    # This would avoid needing the DB downstream in shell_interface, and make the
    # Configuration complete.

    def model_post_init(self, __context):
        """
        Used to specifically update defaults only for specific configs
        """
        # For sequencing, we know we want to fill default time parameters
        # with those from experiment.
        # We are handling similar replacements in shell_parameters in the shell
        # config. However, generalizing templates might be important (see #22)

        sequencing_dict = self.sequencing.model_dump()
        sequencing_dict = replace_curlies_in_dict(sequencing_dict, self.model_dump())
        self.sequencing = Sequencing(**sequencing_dict)

        # after model is validated, instantiate without components
        self.model = Model(
            **self.model.model_dump(exclude={"components"})
        )

        # after experiment is validated, instantiate without components
        self.experiment = Experiment(
            **self.experiment.model_dump(exclude={"components"})
        )

    def get_unique_key_value(self, key: str):
        """Search recursively through the nested dicts of the configuration to try and find a specified key
        and return its value if the key is unique. If mulitple instances of they key exist, return an error.
        """
        # This is searching the whole imsi config for a match for key (variable)
        result = set(list(recursive_lookup(key, self.model_dump())))
        if len(result) != 1:
            # No results or no unique results
            # No unique results is a major challenge. But resolving this would require specifying more information
            # than just {{variable}}. For example, something like "input_files" which appears in the configs for
            # each model would not be unique. So far, we only need to search for uniquely defined values.
            raise ValueError(f"Could not find a unique imsi definition of {key}")
        else:
            return result.pop()


class ConfigManager:
    """This class is used to establish configuration objects, as well as save/load them for later use"""

    # Injecting the DB is good, but it might be useful to initialize it inline.
    def __init__(self, db: ConfigDatabase = None):
        self.db = db

    def create_configuration(
        self,
        model_name: str,
        experiment_name: str,
        machine_name: str = "",
        compiler_name: str = "",
        sequencer_name: str = "",
        flow_name: str = "",
        postproc_profile: str = "",
        **kwargs,
    ):
        """Create the individual instances of config elements and return a configuration composed of these"""
        # Capture all key=value pairs passed, filtering out "self" and kwargs (else it would be nested below a kwargs key)
        setup_params = {
            key: value
            for key, value in locals().items()
            if key != "self" and key != "kwargs"
        }
        # Merge kwargs into setup_params, without being nested
        setup_params.update(kwargs)

        if self.db is None:
            raise RuntimeError("imsi ConfigManager has no database")

        experiment = self.create_experiment(experiment_name, model_name)
        model = self.create_model(model_name)
        MACH_N, machine = self.create_machine(machine_name)
        components = self.create_components(model, experiment)
        COMP_N, compiler = self.create_compiler(machine, compiler_name)
        setup_params = self.create_SetupParams(setup_params, machine)  # improvable
        utilities = self.create_utilities()
        (SEQ_N, FLOW_N), sequencing = self.create_sequencing(
            machine, experiment, model, setup_params.sequencer_name, flow_name
        )
        POST_N, postproc = self.create_postproc(model, experiment, postproc_profile)

        # update the setup_params (obj) with all resolved config
        # component names:
        setup_params.machine_name = MACH_N
        setup_params.postproc_profile = POST_N
        setup_params.sequencer_name = SEQ_N
        setup_params.flow_name = FLOW_N
        setup_params.compiler_name = COMP_N

        return Configuration(
            model=model,
            experiment=experiment,
            components=components,
            machine=machine,
            compiler=compiler,
            postproc=postproc,
            setup_params=setup_params,
            utilities=utilities,
            sequencing=sequencing,
        )

    def create_experiment(self, experiment_name: str, model_name: str) -> Experiment:
        return ExperimentFactory.create_from_database(
            self.db, experiment_name, model_name
        )

    def create_model(self, model_name: str) -> Model:
        model_data = self.db.get_parsed_config("models", model_name)
        return Model(name=model_name, **model_data)

    def create_components(
        self, model: Model, experiment: Experiment
    ) -> Components:
        # deep merge the components of the experiment and model
        # experiment overrides model when params are shared
        components = update(model.components, experiment.components)
        # Create the merged components object
        return Components(**components)

    def create_machine(self, machine_name: str = None) -> Machine:
        return Machine.create_from_database(self.db, machine_name)

    def create_compiler(self, machine: Machine, compiler_name: str = None) -> Compiler:
        compiler_name, cf = CompilerFactory.create_from_database(self.db, machine, compiler_name)
        return compiler_name, cf

    def create_postproc(
        self, model: Model, experiment: Experiment, postproc_profile: str
    ) -> PostProcessing:
        return PostprocFactory.create_from_database(
            self.db, model, experiment, postproc_profile
        )

    def create_SetupParams(self, setup_params: Dict, machine: Machine) -> SetupParams:
        if "sequencer_name" in setup_params:
            if not setup_params["sequencer_name"]:  # It is empty
                setup_params["sequencer_name"] = machine.get_default_sequencer()
        return SetupParams(**setup_params)

    def create_utilities(self) -> Utilities:
        utilities_data = copy.deepcopy(self.db.get_config("utility_config"))
        return Utilities(**utilities_data)

    def create_sequencing(
        self,
        machine: Machine,
        experiment: Experiment,
        model: Model,
        sequencer_name: str,
        flow_name: str,
    ) -> Sequencing:
        return SequencingFactory.create_from_database(
            self.db,
            machine=machine,
            experiment=experiment,
            model=model,
            sequencer_name=sequencer_name,
            flow_name=flow_name,
            model_type=experiment.model_type,
        )

    @classmethod
    def save_configuration(self, configuration: Configuration, filepath: str):
        """Save the configuration to a file"""

        dumped = configuration.model_dump()
        if not dumped:
            raise ValueError("Configuration is empty, nothing to save")

        with open(filepath, 'w') as f:
            # preserve order with sort_keys=False
            yaml.dump(dumped, f, default_flow_style=False, sort_keys=False)

    def load_state(self, filepath) -> Configuration:
        """Load the configuration state from a file"""
        with open(filepath, "rb") as file:
            return pickle.load(file)

    @classmethod
    def save_state(self, configuration: Configuration, filepath: str):
        """Pickle the configuration object"""
        with open(filepath, 'wb') as f:
            pickle.dump(configuration, f)

    # Simple de-serialization
    def load_configuration(self, filepath) -> Configuration:
        """Load the configuration from a file"""
        with open(filepath, 'rb') as f:
            cfg = yaml.safe_load(f)
        return Configuration(**cfg)



"""
imsi CLI
--------

The entry-point console script that interfaces all users commands to imsi.

imsi has several categories of sub-commands. As this module develops further,
the sub-groups are implemented in the relevant downstream modules.
"""

import click
from imsi.cli.sectioned_group import SectionedGroup

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

DECO_HELP="""\b
     ___   _________   _________   ___
    |___| | __   __ | |   ______| |__ |
    | |   | | | | | | |_______ -.   | |
    | |   | | | | | |         | |   | |
 ___| |   | | |_| | |  _______| |   | |___
|___\\_|   |_|     |_| |_________|   |_____|

IMSI CLI — manage configs, builds, runs, and tools.
"""

ECHO_INVOKE_COMMAND_NAMES = [
    'config', 'reload', 'set', 'build', 'submit', 'save-restarts',
    'tapeload-rs', 'get-src'
    ]

@click.group(
    cls=SectionedGroup,
    invoke_without_command=True,
    context_settings=CONTEXT_SETTINGS,
    help=DECO_HELP,
)
@click.version_option(package_name="cimsi")
@click.option("-f", "--force", is_flag=True, help="Force the operation")
@click.pass_context
def cli(ctx, force):
    ctx.ensure_object(dict)
    ctx.obj["FORCE"] = force

    if ctx.invoked_subcommand is None:
        # print help if no subcommand entered
        click.echo(ctx.get_help())
    elif ctx.invoked_subcommand in ECHO_INVOKE_COMMAND_NAMES:
        click.echo(f"IMSI {ctx.invoked_subcommand}")


cli.add_lazy_command("imsi.cli.core_cli.setup")
cli.add_lazy_command("imsi.cli.core_cli.config")
cli.add_lazy_command(
    "imsi.cli.core_cli.save_restarts",
    name="save-restarts",
    context_settings={"ignore_unknown_options": True},
    )
cli.add_lazy_command(
    "imsi.cli.core_cli.tapeload_rs",
    name="tapeload-rs",
    context_settings={"ignore_unknown_options": True},
    add_help_option=False,  # Disable automatic help flag,
    )
cli.add_lazy_command(
    "imsi.cli.core_cli.build",
    short_help="Compile model components.",
    context_settings={"ignore_unknown_options": True},
    add_help_option=False,  # Disable automatic help flag
)
cli.add_lazy_command("imsi.cli.core_cli.submit")
cli.add_lazy_command("imsi.cli.core_cli.status")
cli.add_lazy_command("imsi.cli.core_cli.reload")
cli.add_lazy_command("imsi.cli.core_cli.set")
cli.add_lazy_command("imsi.cli.core_cli.get_src", name='get-src')
cli.add_lazy_command("imsi.cli.core_cli.override")
cli.add_lazy_command("imsi.tools.disk_tools.disk_tools_cli.clean")
cli.add_lazy_command("imsi.tools.list.list_cli.list")
cli.add_lazy_command("imsi.tools.menu.menu_cli.setup_menu", name="setup-menu")
cli.add_lazy_command("imsi.cli.core_cli.log_state", name="log-state")

cli.add_lazy_command("imsi.tools.ensemble.ensemble_cli.ensemble")
cli.add_lazy_command("imsi.tools.validate.validate_cli.validate", name="validate")
cli.add_lazy_command("imsi.tools.time_manager.timer_cli.chunk_manager", name="chunk-manager")
cli.add_lazy_command("imsi.tools.simple_sequencer.iss_cli.iss")



from pydantic import BaseModel, ConfigDict, Field
from imsi.config_manager.schema.types import DateCoerce


class RunDates(BaseModel):
    model_config = ConfigDict(extra='allow')
    run_start_time: DateCoerce = Field(..., description='Start time of the run.')
    run_stop_time: DateCoerce = Field(..., description='End time of the run.')
    run_segment_start_time: DateCoerce = Field(
        ..., description='Start time of the run segment.'
    )
    run_segment_stop_time: DateCoerce = Field(
        ..., description='End time of the run segment.'
    )
    model_chunk_size: str = Field(..., description='Size of the model chunk.')
    model_internal_chunk_size: str = Field(
        ..., description='Size of the model internal chunk.'
    )
    postproc_chunk_size: str = Field(..., description='Size of the postproc chunk.')


class Sequencing(BaseModel):
    """Sequencing configuration dataclass"""

    model_config = ConfigDict(extra='allow')
    run_dates: RunDates




from pathlib import Path
from typing import Iterable
import pandas as pd
from rich.console import Console
from rich.table import Table
from rich.panel import Panel

from imsi.config_manager.config_manager import database_factory
from typing import Callable, Hashable


def _filter_dataframe(
    df: pd.DataFrame,
    filters: dict[Hashable, str | None],
    match_fn: Callable[[pd.Series, str], pd.Series] = lambda col, val: col == val,
) -> pd.DataFrame:
    for column, value in filters.items():
        if value is not None:
            df = df[match_fn(df[column], value)]
    return df


def _load_model_experiment_pairs(imsi_config_path: Path) -> pd.DataFrame:
    experiments_db = database_factory(imsi_config_path / "experiments")
    experiments = experiments_db.get_config("experiments")

    data = []
    for experiment in experiments:
        models = experiments_db.get_parsed_config("experiments", experiment).get("supported_models", [])
        for model in models:
            data.append((model, experiment))

    df = pd.DataFrame(data, columns=["model", "experiment"]) if data else pd.DataFrame()

    return df


def gather_choices_df(
    repo_paths: Iterable[Path],
    repo_sources: Iterable[str],
    relative_imsi_config_path: Path,
    filter_model: str = None,
    filter_experiment: str = None,
) -> pd.DataFrame:

    all_rows = []
    for repo, source in zip(repo_paths, repo_sources):
        imsi_config_path = repo / relative_imsi_config_path
        df = _load_model_experiment_pairs(imsi_config_path)

        if df.empty:
            continue

        df["repo"] = str(repo)
        df["source"] = source

        filters = {
            "model": filter_model,
            "experiment": filter_experiment,
        }
        df = _filter_dataframe(df, filters)

        if not df.empty:
            all_rows.append(df)

    if not all_rows:
        return pd.DataFrame()

    return pd.concat(all_rows, ignore_index=True)


def render_choices_table(df: pd.DataFrame) -> None:
    """
    Render available models/experiments grouped by site/source,
    always showing full model names.
    """
    console = Console()

    if df.empty:
        console.print("[red]No matching configurations found.[/red]")
        return

    grouped_data = []
    for (source, repo), df_group in df.groupby(["source", "repo"]):
        grouped = df_group.groupby("model")["experiment"].apply(
            lambda exps: ", ".join(sorted(set(exps)))
        )
        grouped_data.append(((source, repo), grouped))

    # Render each group
    for (source, repo), grouped in grouped_data:
        table = Table(show_header=True, box=None, padding=(0, 1))

        table.add_column("Model", justify="left", style="magenta", no_wrap=True)
        table.add_column("Supported Experiments", justify="left", style="cyan", overflow="fold")

        for model, experiments in grouped.items():
            table.add_row(model, experiments)

        panel_title = f"[green]{repo}[/green] ({source})"
        console.print(Panel(table, title=panel_title, border_style="cyan"))


def list_all_choices(
    repo_paths: Iterable[Path],
    repo_sources: Iterable[str],
    relative_imsi_config_path: Path,
    filter_model: str = None,
    filter_experiment: str = None,
):
    df = gather_choices_df(
        repo_paths=repo_paths,
        repo_sources=repo_sources,
        relative_imsi_config_path=relative_imsi_config_path,
        filter_model=filter_model,
        filter_experiment=filter_experiment
    )
    render_choices_table(df)



"""
scheduler_tools
================

A module to create submission scripts and tools for different schedulers based on input directives

This is not the core aim of imsi, but is included for development and testing, and in particular it
enables the imsi shell sequencer (:mod:`imsi.iss`) to work across platforms.
"""

import json
import subprocess
from functools import reduce
from imsi.utils.general import get_date_string

fields = [
    ['Job_Name'],
    ['Job_Owner'],
    ['job_state'],
    ['queue'],
    ['Resource_List', 'nodect'],
    ['Resource_List', 'walltime']
]


def pbs_q_query(user=None):
    """
    Not functional
    """
    qstring = subprocess.check_output('qstat -f -Fjson', shell=True)
    qjson = json.loads(qstring)

    for job, job_info in qjson['Jobs'].items():
        pstr=""
        vals = [job]
        for f in fields:
            vals.append(str(reduce(dict.get, f, job_info)).split('@')[0][:65] )

        # Only for running jobs do used times exist
        if vals[3] == 'R':
            f = ['resources_used', 'walltime']
            vals.append(str(reduce(dict.get, f, job_info)).split('@')[0][:65] )
        else:
            vals.append('00:00')

        if user:
            if (job_info['Job_Owner'].split('@')[0] != user):
                #print(job_info['Job_Owner'].split('@')[0], user, job_info['Job_Owner'].split('@')[0]==user)
                continue
        print('{:15s} {:65s} {:8s} {:1s} {:15s} {:4s} {:10s} {:10s}'.format(*vals))

# TODO:
# Make this a general scheduler class. Then have subclass as caps to different schedulers.
# Each class should then have functions to set the resources / directives, etc.
# See: https://realpython.com/python-interface/
class scheduler():
    """
    A class that absracts properties for interacting  with different batch schedulers

    This is at a general level. It could be used to provide generic functionality,
    maybe a bit like rsub or jobsub, but not quite.
    """
    def __init__(self, scheduler_name):
        self.type = scheduler_name
        print("Init scheduler")
        # Move this into a json!(?)
        self._syntax = {
            "pbs" : {
                "directive_prefix" : "#PBS",
                "submission_command" : "qsub",
                "queue_info_command" : "qstat",
                "output_redirect" : '-o "{PATH}.o"',
                "nominal_directives" : [
                    "-l walltime=02:00:00",
                    "-l select=1:mem=100GB",
                    "-j oe"
                ]
            },
            "slurm" : {
                "directive_prefix" : "#SBATCH",
                "submission_command" : "sbatch",
                "queue_info_command" : "squeue",
                "output_redirect" : "-o {PATH}_%j.out",
                "nominal_directives" : [
                    "--time=04:00:00",
                    "--nodes=1"
                ]
            }
        }
        if scheduler_name in self._syntax.keys():
            for key, v in self._syntax[scheduler_name].items():
                setattr(self, key, v)
        else:
            raise ValueError(f"Unsupported scheduler {scheduler_name}. Supported schedulers are {','.join(list(self.syntax.keys()))}")





"""
A collection of 'hooks' (functions) that can be called by imsi.

Mainly designed to be executed with config_hooks_manager.py

Functions are named using the following convention:

    {imsi_step}_{further description}

While naming convention is useful for organization, it is left to the
developer to name and call these functions appropriately.
"""

import os
import subprocess

from imsi.config_manager.config_manager import Configuration
from imsi.shell_interface.shell_config_parameters import generate_shell_parameters
from imsi.utils.general import write_shell_script

def config_postproc_netcdf_conversion(configuration: Configuration, force=False):
    """Configuration for netcdf conversion (ncconv) metadata via pop_nc_tables.

    While netcdf conversion occurs as a post-processing step, configuring
    the metadata required can be done upstream.
    """

    def _get_imsi_config_for_netcdf_conversion():
        # get configuration sections/parameters required for ncconv config

        # from setup_params
        setup_config = {
            "WRK_DIR": configuration.setup_params.work_dir,
            "runid": configuration.setup_params.runid,
        }

        # from model/experiment - required metadata
        wanted_model_keys = ['source_id', 'variant_id']
        wanted_experiment_keys = [
            'activity_id', 'experiment_id', 'parent_branch_time', 'parent_runid', 'subexperiment_id'
            ]

        model_config = dict(
            (k, configuration.model.model_dump()[k])
            for k in wanted_model_keys
        )
        experiment_config = dict(
            (k, configuration.experiment.model_dump()[k])
            for k in wanted_experiment_keys
        )
        # merge model and experiment config
        # into a single dictionary
        modex_config = {**model_config, **experiment_config}

        # from sequencing - time params
        rundt = configuration.sequencing.model_dump()['run_dates']
        seq_config = {
            'run_start_time': rundt['run_start_time'],
            'run_stop_time': rundt['run_stop_time']
        }

        # from postproc netcdf_conversion - processing info / metadata

        ncconv_config = configuration.postproc.netcdf_conversion

        configout = {
            "setup_params": setup_config,
            "modex": modex_config,
            "sequencing": seq_config,
            "postproc": ncconv_config,
        }

        return configout

    SRC_PATH = configuration.setup_params.source_path  # /src -> /CanESM super-repo
    pop_nc_tables_exec = os.path.join(SRC_PATH, 'CCCma_tools', 'scripts', 'comm', 'pop_nc_tables')

    #--------------------
    # get config
    #--------------------

    config_for_ncconv = _get_imsi_config_for_netcdf_conversion()

    # merge all config together
    # **CAUTION** the insertion ORDER MATTERS of the parameters
    # (key-value in ncconv_shell_config) for when the shell file
    # is sourced downstream (ie variable interpolation)
    # FIXME not ideal beacause the string interpolation within imsi config
    # could change and then this could be mangled.
    # For now, at least keep this outside of the subfunction.
    ncconv_shell_config = {'SRC_PATH': SRC_PATH}
    ncconv_shell_config.update(config_for_ncconv['setup_params'])
    ncconv_shell_config.update(config_for_ncconv['modex'])
    ncconv_shell_config.update(config_for_ncconv['sequencing'])
    ncconv_shell_config.update(config_for_ncconv['postproc'])

    # renaming (imsi -> config-canesm)
    # this is required because the pop_nc_tables script expects certain
    # variable names (originally developed for config-canesm tooling)
    aliases = {
        'ncconv_exptab_src': 'ncconv_exptab_repo',
        'ncconv_exptab_commit': 'ncconv_exptab_cmmt',
        'ncconv_commit': 'ncconv_cmmt',
        'ncconv_exptab_commit': 'ncconv_exptab_cmmt'
    }

    ncconv_shell = {}
    for k,v in ncconv_shell_config.items():
        k = aliases[k] if k in aliases else k
        ncconv_shell[k] = v

    force_arg = "true" if force else "false"

    #--------------------
    # write shell file
    #--------------------
    shfile_path = os.path.join(
        configuration.get_unique_key_value('run_config_path'),'ncconv_shell_parameters'
        )

    write_shell_script(shfile_path, generate_shell_parameters(ncconv_shell), mode='w')

    #--------------------
    # call executable
    #--------------------
    # final check for args passed to pop_nc_tables (redundant)
    if not all([os.path.exists(shfile_path), os.path.exists(SRC_PATH)]):
        raise FileNotFoundError("can't execute pop_nc_tables")

    try:
        # pop_nc_tables args:
        #    config  -  config file that containin configuration params (shell)
        #    srcdir  -  source repo location (source code)
        proc = subprocess.run(
            [pop_nc_tables_exec, f'config={shfile_path}', f'srcdir={SRC_PATH}', f'force={force_arg}']
            )
        proc.check_returncode()
    except subprocess.CalledProcessError as e:
        raise ChildProcessError("Failed call: {cmd}\n{err}".format(cmd=' '.join(proc.args), err=proc.stderr))



from typing import Any
from omegaconf import DictConfig, OmegaConf
from typing import List, Union, Dict
import warnings


def convert_to_bracket_notation(key_string: str) -> str:
    """Convert a key string to bracket notation compatible with omegaconf keypaths"""
    keys = key_string.split(":")
    keylist = [f"[{key}]" for key in keys[1:]]
    return keys[0] + "".join(keylist)


def replace_aliases(cfg: DictConfig, aliases: dict) -> DictConfig:
    """Replace top-level keys in cfg with nested aliases (like 'setup:exp')."""
    for alias, target_path in aliases.items():
        target_path = convert_to_bracket_notation(target_path)
        if alias in cfg:
            value = cfg[alias]
            OmegaConf.update(cfg, target_path, value, merge=True)
            del cfg[alias]
    return cfg


def add_nested_key(config_dict: DictConfig, key: str, value: Any):
    """
    Traverse key with ':' to add it to a nested dictionary structure.
    """
    try:
        key_path_full = convert_to_bracket_notation(key)
        OmegaConf.update(
            config_dict,
            key_path_full,
            value,
            force_add=True,
            merge=True,
        )
    except Exception as e:
        raise ValueError(
            f"Error adding key {key} to config_dict. This may be caused by conflicting definitions of key paths provided as both a parameter and nested key."
        ) from e


def get_keys(member: DictConfig, prefix="") -> set:
    """Recursively extract all keys from a DictConfig or dictionary, including nested ones."""
    keys = set()
    for key, value in member.items():
        full_key = f"{prefix}:{key}" if prefix else key
        keys.add(full_key)
        if isinstance(value, (DictConfig, dict)):
            keys.update(get_keys(value, full_key))
    return keys


def warn_on_overlapping_keys(cfg1: DictConfig, cfg2: DictConfig):
    """Warn if any keypaths overlap between two configs."""
    keys1 = get_keys(cfg1)
    keys2 = get_keys(cfg2)
    overlap = keys1 & keys2
    if overlap:
        warnings.warn(
            colour(
                f"Overlapping keys detected between table and entry config: {sorted(overlap)}. The values in the table will override those in the entry config.",
                "yellow"
            ),
            UserWarning
        )


def colour(text, c):
    """Basic ANSI colourizer."""
    codes = dict(red="\033[91m", green="\033[92m", yellow="\033[93m", reset="\033[0m")
    return f"{codes[c]}{text}{codes['reset']}"


def check_keys_against_reference(dict_list: List[Union[DictConfig, dict]]):
    """Compare each config to the first one and report any key mismatches."""
    if not dict_list:
        print(colour("⚠️  Config list is empty — skipping key comparison.", "yellow"))
        return

    ref_keys = get_keys(dict_list[0])

    for i, member in enumerate(dict_list[1:], start=1):
        member_keys = get_keys(member)
        extra = member_keys - ref_keys
        missing = ref_keys - member_keys

        if extra or missing:
            print(colour(f"\n  Key mismatch between config[0] (reference) and config[{i}]:", "yellow"))
            print(colour("\n  This message can be safely ignored if this is intentional.", "yellow"))
            if extra:
                print(colour(f"  - Extra keys in member[{i}]:", "green"))
                for key in sorted(extra):
                    print(f"    + {key}")
            if missing:
                print(colour(f"  - Missing keys in member[{i}]:", "red"))
                for key in sorted(missing):
                    print(f"    - {key}")
        else:
            print(colour(f"✓ Member[{i}] keys exactly match reference keys.", "green"))


def remove_explicit_none(member):
    """Remove keys with explicit None values recursively from config_list"""
    for k, v in list(member.items()):
        if v is None:
            member.pop(k)
        if isinstance(v, (dict, DictConfig)):
            remove_explicit_none(v)
    return OmegaConf.create(member)


def validate_runids(cfg_list: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Validate or auto-generate unique 'runid' values in each configuration's 'setup' block."""

    # Check and collect runids
    runid_list = []
    for idx, member in enumerate(cfg_list):
        try:
            runid = member['setup']['runid']
            runid_list.append(runid)
        except (KeyError, TypeError):
            raise ValueError(
                colour(
                    f"\n❌ Missing 'setup.runid' in member index {idx}. Each member must have a 'setup' dictionary with a 'runid' key.",
                    "red"
                )
            )

    # Check uniqueness
    unique_runids = set(runid_list)
    if len(unique_runids) < len(cfg_list):
        warnings.warn(
            colour(
                f"\n⚠️  Duplicate 'runid' values detected across {len(cfg_list)} ensemble members "
                f"(only {len(unique_runids)} unique): {runid_list}\n"
                "➡️  To ensure consistent file naming and output isolation, each runid will be auto-suffixed with its index (e.g., 'runid-0', 'runid-1', ...).",
                "yellow"
            )
        )

        for i, member in enumerate(cfg_list):
            old_runid = member['setup']['runid']
            new_runid = f"{old_runid}-{i}"
            cfg_list[i]['setup']['runid'] = new_runid

    return cfg_list


def validate_table_data(table, show_diffs=False):
    table = [remove_explicit_none(d) for d in table]
    table = validate_runids(table)
    if show_diffs:
        check_keys_against_reference(table)

    return table



from imsi.tools.ensemble.table_utils.data_model import (
    EnsembleLevelVars,
    MemberLevelVars,
)
from imsi.tools.ensemble.table_utils.table_utils import (
    add_nested_key,
    colour,
    validate_table_data,
    warn_on_overlapping_keys,
)

from pydantic import BaseModel, FilePath
from omegaconf import DictConfig, OmegaConf
from typing import Any, Dict, List, Tuple, Optional
import copy
import pandas as pd
import warnings

warnings.simplefilter("always", DeprecationWarning)


def validate_list_lengths(member_level_size: int, table_length: int):
    """Ensure all list values in a flat dictionary have the same length."""

    # check that broadcasted_members and table have the same length
    if member_level_size is not None and member_level_size != table_length:
        raise ValueError(
            colour(
                f"\n❌  The number of broadcasted members ({member_level_size}) does not match the number of table entries ({table_length}). ",
                "red"
            )
        )


class YAMLVars(BaseModel):
    member_level: MemberLevelVars

    def get_table(self, show_diffs):
        broadcasted_members = BroadcastedConfigList(member_level=self.member_level)
        cfg_list = broadcasted_members.expand_config()
        cfg_list = validate_table_data(cfg_list)
        return cfg_list


class CSVTableVars(BaseModel):
    """Class with supporting functions to read and process an ensemble text file."""

    ensemble_level: EnsembleLevelVars
    member_level: MemberLevelVars

    def read_table(self) -> DictConfig:
        """Load csv as pandas dataframe and inject heirarchy into configuration object"""
        data = pd.read_csv(
            self.ensemble_level.config_table,
            comment="#",
            sep=r"\s*,\s*",
            engine="python",
            dtype_backend="pyarrow"  # preserves atual integer types instead of converting to float
        )
        # replace missing values to be explicit None
        data = data.replace({float("nan"): None})
        data_dict = OmegaConf.create({col: data[col].tolist() for col in data.columns})
        config_dict = OmegaConf.create({})

        for key in data_dict.keys():
            if ":" in key:
                # traverse key to add to nested dictionary
                try:
                    add_nested_key(config_dict, key, data_dict[key])
                except Exception:
                    raise ValueError(
                        f"Error adding key {key} to data_dict. This is likely caused by conflicting definitions of key paths provided as both a parameter and nested key."
                    )
            else:
                # Directly assign the list if it's not a nested key
                config_dict[key] = data_dict[key]
        
        return config_dict

    def get_table(self, show_diffs):
        """Combine the csv and member level config data to produce a list of configurations."""
        # create and resulve the table
        table = self.read_table()
        table_length = len(list(MemberLevelVars.get_listed_vars(table).values())[0])
        # update the member level data with the table data
        validate_list_lengths(self.member_level.ensemble_size, table_length)

        warn_on_overlapping_keys(
            self.member_level.config_data, table
        )

        member_data = OmegaConf.merge(self.member_level.config_data, table)
        # expand to list of configurations
        cfg_list = BroadcastedConfigList(
            member_level=MemberLevelVars(config_data=member_data)
        ).expand_config(ensemble_size=table_length)

        # remove any keys with explicit None values
        cfg_list = validate_table_data(cfg_list, show_diffs)

        return cfg_list


class TextTableVars(CSVTableVars):

    """Inherit same functionality from CSVTableVars but with different read_table method."""

    def read_table(self):
        """Load text table as pandas dataframe and inject heirarchy into configuration object"""

        warnings.warn(
            colour(
                "\n⚠️  Text files are a legacy ensemble tool input and will be deprecated. Please provide a csv table instead.",
                "yellow"
            ),
            DeprecationWarning,
        )

        data = pd.read_csv(
            self.ensemble_level.config_table,
            sep=r"\s+",
            comment="#",
        )
        data_dict = {col: data[col].tolist() for col in data.columns}
        data_dict = OmegaConf.create(data_dict)
        config_dict = OmegaConf.create({})

        for key in data_dict.keys():
            if ":" in key:
                add_nested_key(config_dict, key, data_dict[key])
            else:
                # Directly assign the list if it's not a nested key
                config_dict[key] = data_dict[key]

        return config_dict


class BroadcastedConfigList(BaseModel):
    """Takes information from the broadcasted vars and expands into a list of configurations"""

    member_level: MemberLevelVars

    def _set_nested_value(
        self, data: Dict[str, Any], keys: Tuple[str, ...], value: Any
    ):
        """Recursively set a value in a nested dictionary."""
        for key in keys[:-1]:
            data = data[key]
        data[keys[-1]] = value

    def expand_config(self, ensemble_size=None) -> List[Dict[str, Any]]:
        """Expand the configuration to produce a list of configurations, one for each ensemble member."""

        if not self.member_level.config_data:
            return []
        
        ensemble_size = self.member_level.ensemble_size if ensemble_size is None else ensemble_size
        base_config = self.member_level.config_data
        listed_vars = MemberLevelVars.get_listed_vars(base_config)

        config_list = []
        for i in range(ensemble_size):
            member_config = copy.deepcopy(base_config)
            for keys, value in listed_vars.items():
                self._set_nested_value(member_config, keys, value[i])
            config_list.append(member_config)

        return OmegaConf.create(config_list)


class YAMLTableVars(BaseModel):
    member_level: MemberLevelVars
    config_table: Optional[FilePath] = None

    def get_table(
        self,
        show_diffs
    ):
        if self.config_table is not None:
            table = OmegaConf.load(self.config_table)
        else:
            raise ValueError("No config table provided.")

        validate_list_lengths(self.member_level.ensemble_size, len(table))

        # create a broadcasted config list from the member level data
        broadcasted_members = BroadcastedConfigList(
            member_level=MemberLevelVars(config_data=self.member_level.config_data)
        ).expand_config(ensemble_size=len(table))

        if broadcasted_members:
            for i in range(len(table)):
                warn_on_overlapping_keys(
                    broadcasted_members[i], table[i]
                )
                table[i] = OmegaConf.merge(
                    broadcasted_members[i], table[i]
                )

        table = validate_table_data(table, show_diffs=show_diffs)

        return table



import click
from pydantic import ValidationError
from rich.prompt import Prompt



@click.command(
    name="setup-menu",
    short_help="Interactive prompt to populate an imsi setup command.",
    help="Interactive prompt to populate an imsi setup command for on-disk repositories.",
    epilog="""On-disk respositories are set by the .rc files.
    Check the .rc sources using imsi list.
    """
    )
@click.option('--repo-path', default=None,
              help="Path to a repo or directory of repos.")
@click.option('-y', '--execute', is_flag=True,
              help="Execute the setup command after selection.")
def setup_menu(repo_path, execute):
    from imsi.user_interface.setup_manager import ValidatedSetupOptions, InvalidSetupConfig
    from imsi.tools.menu.menu_helpers import (
        select_imsi_config_with_questionary,
        prompt_additional_options,
        build_setup_command,
        print_command,
        execute_command,
    )
    from imsi.tools.list.list_cli import get_repo_paths

    repo_paths, repo_sources, config_path = get_repo_paths(repo_path)

    while True:
        # Step 1: Prompt for selection
        selection = select_imsi_config_with_questionary(repo_paths, repo_sources, config_path)
        if not selection:
            print("\033[1;31m[✗] No item selected\033[0m")
            return

        # Step 2: Prompt for CLI extras
        additions = prompt_additional_options()
        if additions is None:
            return
        else:
            selection.update(additions)

        # Step 3: Validate
        try:
            ValidatedSetupOptions(**selection)
            break  # valid config, exit loop
        except (InvalidSetupConfig, ValidationError) as e:
            print(f"\033[1;31m[✗] Invalid setup options: {e}\033[0m")
            if Prompt.ask("Try again?", choices=["yes", "no"], default="yes") == "no":
                print("\033[2m[Exiting setup]\033[0m")
                return

    # Step 4: Build and show setup command
    setup_cmd = build_setup_command(selection)
    print_command(setup_cmd)

    # Step 5: Ask if it should be executed
    if execute or Prompt.ask("Execute now?", choices=["yes", "no"], default="yes") == "yes":
        execute_command(setup_cmd)


import pytest
from imsi.tools.ensemble.table_utils.data_model import MemberLevelVars
from omegaconf import OmegaConf

from pathlib import Path

parent_dir = Path(__file__).resolve().parent


@pytest.fixture
def config_params(request):
    pars = OmegaConf.create({
        "repo": request.config.getoption("--target-repo"),
        "branch": request.config.getoption("--target-branch"),
        "model": request.config.getoption("--target-model"),
        "experiment": request.config.getoption("--target-exp")
    })
    return pars


@pytest.fixture
def setup_block(config_params):
    return OmegaConf.create({
        "ver": config_params["branch"],
        "repo": config_params["repo"],
        "fetch_method": "clone",
        "machine": "ppp6",
        "model": config_params["model"],
    })


@pytest.fixture
def config_data_one(setup_block):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_block,
                "runid": ["run1"],
            }
        }
    })


@pytest.fixture
def config_bash_resolve(setup_block):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_block,
                "runid": ["run1"],
            },
            "some_path": "/this/is/a/path/\\${member_level.setup.runid}",
        }
    })


@pytest.fixture
def config_data_two(setup_block):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_block,
                "runid": ["run1", "run2"],
            }
        }
    })


@pytest.fixture
def good_multiple_lists(setup_block):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_block,
                "runid": ["run1", "run2"],
            },
            "param": {"param1": [1, 2]},
            "another_param": {"another_another_param": [1, 2]},
        }
    })


@pytest.fixture
def bad_multiple_lists(setup_block):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_block,
                "runid": ["run1", "run2"],
            },
            "param": {"good_param": [1, 2]},
            "another_param": {"another_bad_param": [1]},
        }
    })


@pytest.fixture
def config_data_no_lists(setup_block):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_block,
                "runid": "run1",
            }
        }
    })


@pytest.mark.parametrize("config_data,expected", [
    ("config_data_one", 1),
    ("config_data_two", 2),
])
def test_ensemble_size_property(request, config_data, expected):
    member_level = MemberLevelVars(config_data=request.getfixturevalue(config_data))
    assert member_level.ensemble_size == expected


@pytest.mark.parametrize("config_data,expected", [
    ("good_multiple_lists", None),
    ("bad_multiple_lists", ValueError),
])
def test_check_list_size_equal(request, config_data, expected):
    member_level = MemberLevelVars(config_data=request.getfixturevalue(config_data))
    if expected:
        with pytest.raises(expected):
            member_level.get_listed_vars(member_level.config_data)
    else:
        member_level.get_listed_vars(member_level.config_data)


@pytest.mark.parametrize("config_data,extracted_list", [
    ("config_data_one", {("member_level", "setup", "runid"): ["run1"]}),
    ("config_data_two", {("member_level", "setup", "runid"): ["run1", "run2"]}),
    ("good_multiple_lists", {
        ("member_level", "setup", "runid"): ["run1", "run2"],
        ("member_level", "param", "param1"): [1, 2],
        ("member_level", "another_param", "another_another_param"): [1, 2],
    }),
])
def test_get_listed_vars(request, config_data, extracted_list):
    member_level = MemberLevelVars(config_data=request.getfixturevalue(config_data))
    assert member_level.get_listed_vars(member_level.config_data) == extracted_list


def test_no_lists(config_data_no_lists):
    member_level = MemberLevelVars(config_data=config_data_no_lists)
    assert member_level.ensemble_size is None


def test_bash_resolve(config_bash_resolve):
    member_level = MemberLevelVars(config_data=config_bash_resolve)
    assert member_level.config_data.member_level.some_path == "/this/is/a/path/${member_level.setup.runid}"


import pytest
from click.testing import CliRunner
from imsi.cli.entry import cli
from pathlib import Path
import subprocess
import shutil
import os
import yaml
from imsi import __version__


from imsi.user_interface.setup_manager import InvalidSetupConfig


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def local_test_dir(request):
    """Create a temporary test directory under the current working directory and clean up after."""
    tmp_dir = Path.cwd() / f"test_tmp_{request.node.name}"
    if tmp_dir.exists():
        shutil.rmtree(tmp_dir)
    tmp_dir.mkdir(parents=True)
    yield tmp_dir
    shutil.rmtree(tmp_dir, ignore_errors=True)


@pytest.mark.parametrize(
    ("ctx_args", "before_args", "delimiter", "after_args"),
    [
        ("-f", None, None, "-c"),
        (None, "-c", '--', "-h"),
        ("-f", "-c", '--', '-c'),
    ]
)
def test_build_fails_when_args_malformed_with_eoo_delimiter(
        runner, local_test_dir,
        ctx_args, before_args, delimiter, after_args
        ):
    cwd_before = os.getcwd()
    os.chdir(local_test_dir)

    build_args = [ctx_args, "build", before_args, delimiter, after_args]
    build_args = [x for x in build_args if x is not None]
    try:
        result = runner.invoke(cli, build_args)
    finally:
        os.chdir(cwd_before)
    assert result.exit_code != 0



import pytest
from pydantic import ValidationError
from imsi.config_manager.schema.experiment import Experiment


@pytest.fixture
def valid_experiment_data():
    """Fixture to provide valid experiment data."""
    return {
        'name': 'Test Experiment',
        'experiment_id': 'exp_001',
        'subexperiment_id': 'sub_001',
        'activity_id': 'act_001',
        'mip_era': 'CMIP6',
        'model_type': 'AOGCM',
        'start_time': '2025',
        'end_time': '2025',
        'parent_runid': 'parent_123',
        'parent_branch_time': '2024',
        'inherits_from': None,
        'supported_models': ['ModelA', 'ModelB'],
        'components': {'atmosphere': 'AMIP', 'ocean': 'OMIP'},
    }


def test_missing_required_fields():
    """Ensure required fields cannot be omitted."""
    with pytest.raises(ValidationError):
        Experiment(name='Test Experiment')


def test_datecoerce_validation(valid_experiment_data):
    """Ensure start_time and end_time accept valid date-like values."""
    valid_experiment_data['start_time'] = 2025
    valid_experiment_data['end_time'] = '2026'
    experiment = Experiment(**valid_experiment_data)
    assert isinstance(experiment.start_time, str)


def test_invalid_supported_model(valid_experiment_data):
    """Ensure error is raised if model_name is not in supported_models."""
    valid_experiment_data['supported_models'] = ['ModelC', 1234]
    with pytest.raises(ValueError):
        Experiment(**valid_experiment_data)  # Not in supported_models


def test_check_supported_models(valid_experiment_data):
    """Ensure check_supported_models validator works."""
    model_name = 'ModelC'
    with pytest.raises(ValueError):
        Experiment(**valid_experiment_data).validate_model_name(model_name=model_name)


def test_optional_fields(valid_experiment_data):
    """Ensure optional fields can be omitted or set to None."""
    valid_experiment_data.pop('inherits_from')
    valid_experiment_data['components'] = None
    experiment = Experiment(**valid_experiment_data)
    assert experiment.inherits_from is None
    assert experiment.components is None  # Should default to None


def test_missing_start_time(valid_experiment_data):
    """Ensure missing start_time field triggers an error."""
    valid_experiment_data.pop('start_time')
    with pytest.raises(ValidationError):
        Experiment(**valid_experiment_data)



import click
import importlib


class LazyCommand(click.Command):
    def __init__(
        self,
        import_path,
        name=None,
        short_help="",
        context_settings=None,
        add_help_option=False,
    ):
        self._import_path = import_path
        self._real_command = None
        self._short_help = short_help
        super().__init__(
            name or import_path.split(".")[-1],
            short_help=short_help,
            context_settings=context_settings or {},
            add_help_option=add_help_option,
        )

    def _load_command(self):
        if self._real_command is None:
            module_path, func_name = self._import_path.rsplit(".", 1)
            module = importlib.import_module(module_path)
            self._real_command = getattr(module, func_name)
        return self._real_command

    def invoke(self, ctx):
        return self._load_command().invoke(ctx)

    def get_help(self, ctx):
        return self._load_command().get_help(ctx)

    def get_params(self, ctx):
        return self._load_command().get_params(ctx)

    def format_help(self, ctx, formatter):
        return self._load_command().format_help(ctx, formatter)

    def format_usage(self, ctx, formatter):
        return self._load_command().format_usage(ctx, formatter)

    def get_short_help_str(self, limit=45):
        return self._short_help or self._load_command().get_short_help_str(limit)


class LazyGroup(click.Group):
    def __init__(self, *args, **kwargs):
        self._lazy_commands = {}
        super().__init__(*args, **kwargs)

    def add_lazy_command(
        self,
        import_path: str,
        name: str | None = None,
        short_help: str = "",
        **lazy_kwargs,
    ):
        """
        Register a lazy‑loaded command or group.

        Parameters
        ----------
        import_path : str
            Dotted import path to a click.Command or click.Group object.
        name : str, optional
            Override the command name shown in the CLI.
        short_help : str, optional
            One‑liner shown in the parent help without importing the module.
        **lazy_kwargs
            Extra keyword args forwarded to LazyCommand
            (e.g. context_settings, add_help_option).
        """
        name = name or import_path.split(".")[-1]
        self._lazy_commands[name] = (import_path, short_help, lazy_kwargs)

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._lazy_commands:
            import_path, short_help, extra = self._lazy_commands[cmd_name]
            module_path, func_name = import_path.rsplit(".", 1)
            module = importlib.import_module(module_path)
            real_cmd = getattr(module, func_name)

            # If it's a Group, return it as‑is (no wrapper needed)
            if isinstance(real_cmd, click.Group):
                return real_cmd

            # Else wrap it, forwarding all extra kwargs
            return LazyCommand(
                import_path,
                name=cmd_name,
                short_help=short_help,
                **extra,
            )

        return super().get_command(ctx, cmd_name)

    def list_commands(self, ctx):
        cmds = list(self.commands.keys())
        lazies = [name for name in self._lazy_commands.keys() if name not in self.commands]
        return cmds + lazies



'''
This is a module containing utilities for handling cftimes, and particularly
generating cftimes from iso8061 strings, and assiting in creating cftime_ranges
appropriate for model chunking.

Various functions here are borrowed, and in cases adapated from xarray, which
itself inherits from pandas (where noted below).
'''

import cftime
import re
from datetime import timedelta
from typing import Tuple

# Possibly replacable with xarray.coding.to_cftime_datetime, but that does not have start/end logic.
def parse_time(time_str: str, cftime_calendar:cftime.datetime, date_type: str = 'start') -> cftime.datetime:
    """
    Parse a time string in iso8601 format and return a cftime date-time object
    for the specified calendar type and date_type.

    Args:
        time_str (str): The time string to parse (iso8601_).
        cftime_calendar (cftime.Datetime): cftime calendar object (e.g. cftime.DatetimeNoleap)
        date_type (str): Specifies if the date is a 'start' or 'end' date (default is 'start').
    Returns:
        (cftime.datetime): The parsed date-time object for the specified calendar.

    If `date_type=start`, any date parameters not specified get set to the beginning
    of the period. If `date_type=end`, any date parameters not specified get set to
    the end of the period.

    Examples:
        parse_time('2016-01', cftime_calendar=cftime.DatetimeNoLeap, date_type='start)
            cftime.DatetimeNoLeap(2016, 1, 1, 0, 0, 0, 0, has_year_zero=True)

        parse_time('2016-01', cftime_calendar=cftime.DatetimeNoLeap, date_type='end)
            cftime.DatetimeNoLeap(2016, 1, 31, 23, 59, 59, 999999, has_year_zero=True)

        parse_time('2016-01-15 16:00', cftime_calendar=cftime.DatetimeNoLeap)
            cftime.DatetimeNoLeap(2016, 1, 15, 16, 0, 0, 0, has_year_zero=True)
    """
    if not time_str:
        raise ValueError("Time string must be provided")
    parsed, resolution = _parse_iso8601_with_reso(cftime_calendar, time_str)
    start, end = _parsed_string_to_bounds(cftime_calendar, resolution, parsed)
    dates = {'start': start, 'end': end}
    try:
        return dates[date_type]
    except KeyError:
        raise ValueError(f'date_type must be one of "start" or "end". Received: {date_type}')


# Part of the hack below
def parse_frequency(freq) -> Tuple[int, str]:
    '''split a frequency string into the multiplier and the
    base frequency (e.g. P12MS n=12, base_freq=MS)
    '''
    #below, matches strings like P1MS, P1Y, PT6H
    #T is optional
    #capture group 1: P or PT; group 2: integer; group 3: time unit
    match = re.match(r"P(T?)(\d+)?(\D+)", freq)
    if not match:
        raise ValueError(f"Frequency '{freq}' is not supported.")

    multiplier = int(match.group(2))
    base_freq = match.group(1) + match.group(3)
    return multiplier, base_freq


# This is a hack, and better replaced with xarray logic
# just have to figure out which functions
def adjust_start_date(start, duration) -> cftime.datetime:
    '''Given a start date with a specified resolution (precision)
       and a frequency (perhaps better duration), adjust the start
       so that it corresponds to the beginning of the "resolution"
       specified in frequency.

       Args:
            start (cftime) : a cftime instance
            duration (str) : the time duration/frequency

       Examples:
       cftime_utils.adjust_start_date(cftime.DatetimeNoLeap(2016,1,14), '12MS')
        cftime.DatetimeNoLeap(2016, 1, 1, 0, 0, 0, 0, has_year_zero=True)

        -> notice the resulting cftime starts on Jan 1, whereas the input time
           started on Jan 14.

        cftime_utils.adjust_start_date(cftime.DatetimeNoLeap(2016,1,14,16,00), '12MS')
            cftime.DatetimeNoLeap(2016, 1, 1, 0, 0, 0, 0, has_year_zero=True)

        cftime_utils.adjust_start_date(cftime.DatetimeNoLeap(2016,1,14,16,00), '1D')
            cftime.DatetimeNoLeap(2016, 1, 14, 0, 0, 0, 0, has_year_zero=True)
    '''

    base_freq = duration.unit

    #some of the units follow xarray conventions, and some also are deprecated in xarray
    date_attrs = {
        'MS': {'day': 1, 'hour': 0, 'minute': 0, 'second': 0},
        'YS': {'month': 1, 'day': 1, 'hour': 0, 'minute': 0, 'second': 0},
        'DS': {'hour': 0, 'minute': 0, 'second': 0},
        'THS': {'minute': 0, 'second': 0},
        'TMS': {'second': 0},
        'TSS': {},
    }

    try:
        return start.replace(**date_attrs[base_freq])
    except KeyError:
        raise ValueError(f"Frequency '{base_freq}' is not supported.")

# Function below are borrowed from xarray, specifically
# xarray.coding.cftime_range, cftimeindex, times
# In some cases functions were slightly modified for sensible usage here

# For reference, this is the license that appears in xarray/pandas source:
# (c) 2011-2012, Lambda Foundry, Inc. and PyData Development Team
# All rights reserved.

# Copyright (c) 2008-2011 AQR Capital Management, LLC
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

#     * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.

#     * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.

#     * Neither the name of the copyright holder nor the names of any
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# standard calendars recognized by cftime
_STANDARD_CALENDARS = {"standard", "gregorian", "proleptic_gregorian"}
def _is_standard_calendar(calendar: str) -> bool:
    return calendar.lower() in _STANDARD_CALENDARS

def get_date_type(calendar, use_cftime=True):
    """Return the cftime date type for a given calendar name."""
    # NCS - not applicable here
    #if cftime is None:
    #    raise ImportError("cftime is required for dates with non-standard calendars")
    #else:
    #    if _is_standard_calendar(calendar) and not use_cftime:
    #        return _nanosecond_precision_timestamp

    calendars = {
        "noleap": cftime.DatetimeNoLeap,
        "360_day": cftime.Datetime360Day,
        "365_day": cftime.DatetimeNoLeap,
        "366_day": cftime.DatetimeAllLeap,
        "gregorian": cftime.DatetimeGregorian,
        "proleptic_gregorian": cftime.DatetimeProlepticGregorian,
        "julian": cftime.DatetimeJulian,
        "all_leap": cftime.DatetimeAllLeap,
        "standard": cftime.DatetimeGregorian,
    }
    # NCS modified for better error handling
    try:
        return calendars[calendar.lower()]
    except KeyError:
        raise ValueError(f"Unsupported calendar type: {calendar}")


def named(name, pattern):
    return "(?P<" + name + ">" + pattern + ")"

def optional(x):
    return "(?:" + x + ")?"

def trailing_optional(xs):
    if not xs:
        return ""
    return xs[0] + optional(trailing_optional(xs[1:]))

def build_pattern(date_sep=r"\-", datetime_sep=r"T", time_sep=r"\:"):
    pieces = [
        (None, "year", r"\d{4}"),
        (date_sep, "month", r"\d{2}"),
        (date_sep, "day", r"\d{2}"),
        (datetime_sep, "hour", r"\d{2}"),
        (time_sep, "minute", r"\d{2}"),
        (time_sep, "second", r"\d{2}"),
    ]
    pattern_list = []
    for sep, name, sub_pattern in pieces:
        pattern_list.append((sep if sep else "") + named(name, sub_pattern))
        # TODO: allow timezone offsets?
    return "^" + trailing_optional(pattern_list) + "$"


_BASIC_PATTERN = build_pattern(date_sep="", time_sep="")
_EXTENDED_PATTERN = build_pattern()
_CFTIME_PATTERN = build_pattern(datetime_sep=" ")
_PATTERNS = [_BASIC_PATTERN, _EXTENDED_PATTERN, _CFTIME_PATTERN]

def parse_iso8601_like(datetime_string):
    '''
    Given a time string, return a dict, with keys of
    year, month, day, hour, minute, second, and the values
    being the integer date value, or None if it was not
    specified in the input string
    (NCS inference, undocumented in xarray).
    '''
    for pattern in _PATTERNS:
        match = re.match(pattern, datetime_string)
        if match:
            return match.groupdict()
    raise ValueError(
        f"no ISO-8601 or cftime-string-like match for string: {datetime_string}"
    )

def _parse_iso8601_with_reso(date_type, timestr):
    '''
    Parses an iso8061 datestring, and returns a cftime instance
    (of date_type) and a specification of the resolution of
    the date string (i.e. specified down to year, month, day...)

    NCS inference - xarray version undocumented.
    '''
    if cftime is None:
        raise ModuleNotFoundError("No module named 'cftime'")

    default = date_type(1, 1, 1)
    result = parse_iso8601_like(timestr)
    replace = {}

    for attr in ["year", "month", "day", "hour", "minute", "second"]:
        value = result.get(attr, None)
        if value is not None:
            # Note ISO8601 conventions allow for fractional seconds.
            # TODO: Consider adding support for sub-second resolution?
            replace[attr] = int(value)
            resolution = attr
    return default.replace(**replace), resolution

# From xarray
def _parsed_string_to_bounds(date_type, resolution, parsed):
    """Generalization of
    pandas.tseries.index.DatetimeIndex._parsed_string_to_bounds
    for use with non-standard calendars and cftime.datetime
    objects.
    """
    if resolution == "year":
        return (
            date_type(parsed.year, 1, 1),
            date_type(parsed.year + 1, 1, 1) - timedelta(seconds=1),
        )
    elif resolution == "month":
        if parsed.month == 12:
            end = date_type(parsed.year + 1, 1, 1) - timedelta(seconds=1)
        else:
            end = date_type(parsed.year, parsed.month + 1, 1) - timedelta(
                seconds=1
            )
        return date_type(parsed.year, parsed.month, 1), end
    elif resolution == "day":
        start = date_type(parsed.year, parsed.month, parsed.day)
        return start, start + timedelta(days=1, seconds=-1)
    elif resolution == "hour":
        start = date_type(parsed.year, parsed.month, parsed.day, parsed.hour)
        return start, start + timedelta(hours=1, seconds=-1)
    elif resolution == "minute":
        start = date_type(
            parsed.year, parsed.month, parsed.day, parsed.hour, parsed.minute
        )
        return start, start + timedelta(minutes=1, seconds=-1)
    elif resolution == "second":
        start = date_type(
            parsed.year,
            parsed.month,
            parsed.day,
            parsed.hour,
            parsed.minute,
            parsed.second,
        )
        #currently we don't support sub-second units, so start = end when specifying down to the second
        return start, start
    else:
        raise KeyError




import unittest
from unittest.mock import Mock
from config_manager import (
    Experiment, Model, Machine, JsonConfigDatabase,
    MachineFactory, Configuration, ConfigManager
)

class TestExperiment(unittest.TestCase):
    def test_experiment_creation(self):
        parameters = {'param1': 1, 'param2': 2}
        exp = Experiment(name='test_exp', parameters=parameters)
        self.assertEqual(exp.name, 'test_exp')
        self.assertEqual(exp.parameters, parameters)

class TestModel(unittest.TestCase):
    def test_model_creation(self):
        parameters = {'param1': 1, 'param2': 2}
        model = Model(name='test_model', parameters=parameters)
        self.assertEqual(model.name, 'test_model')
        self.assertEqual(model.parameters, parameters)

class TestMachine(unittest.TestCase):
    def test_machine_creation(self):
        parameters = {'param1': 1, 'param2': 2}
        machine = Machine(name='test_machine', parameters=parameters)
        self.assertEqual(machine.name, 'test_machine')
        self.assertEqual(machine.parameters, parameters)

class TestJsonConfigDatabase(unittest.TestCase):
    def setUp(self):
        self.mock_combine_configs = Mock(return_value={'mock_config': {'param1': 1}})

    def test_load_config(self):
        with unittest.mock.patch('config_manager.combine_json_configs', self.mock_combine_configs):
            json_db = JsonConfigDatabase(imsi_config_path='/mock/path')
            self.assertEqual(json_db.database, {'mock_config': {'param1': 1}})

    # Add more tests for other methods as needed

#class TestMachineFactory(unittest.TestCase):
#    # Write tests for create_from_database, get_machine_name_from_hostname, find_matching_machine, get_hostname
#
#class TestConfiguration(unittest.TestCase):
#    # Write tests for __init__
#
#class TestConfigManager(unittest.TestCase):
#    # Write tests for __init__, create_configuration, create_experiment, create_model, create_machine

if __name__ == '__main__':
    unittest.main()



import pytest
from pathlib import Path

from pydantic import ValidationError
from imsi.tools.ensemble.table_utils.table_model import CSVTableVars
from imsi.tools.ensemble.table_utils.data_model import (
    MemberLevelVars,
    EnsembleLevelVars,
)
from imsi.tools.ensemble.config import load_config

from omegaconf import OmegaConf
from omegaconf.errors import ConfigTypeError

parent_dir = Path(__file__).resolve().parent


@pytest.fixture
def config_params(request):
    pars = {
        "repo": request.config.getoption("--target-repo"),
        "branch": request.config.getoption("--target-branch"),
        "model": request.config.getoption("--target-model"),
        "experiment": request.config.getoption("--target-exp")
    }
    return pars


@pytest.fixture
def setup_config(config_params):
    return OmegaConf.create({
        "ver": config_params["branch"],
        "repo": config_params["repo"],
        "fetch_method": "clone",
        "machine": "ppp6",
        "model": config_params["model"],
    })


@pytest.fixture
def config_data_one(setup_config):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_config,
                "runid": ["run1"],
            }
        }
    })


@pytest.fixture
def config_data_two(setup_config):
    return OmegaConf.create({
        "member_level": {
            "setup": {
                **setup_config,
                "runid": ["run1", "run2"],
            }
        }
    })


@pytest.fixture
def ensemble_config():
    return OmegaConf.create({
        "user": "test_user",
        "run_directory": str(parent_dir / "./"),
        "config_table": str(parent_dir / "config.yaml"),
        "share_repo": True,
    })


@pytest.fixture
def ensemble_config_txt_one():
    return {
        "user": "test_user",
        "run_directory": str(parent_dir / "./"),
        "config_table": str(parent_dir / "test_config.csv"),
        "aliases": {"alias1": "banana"},
        "share_repo": True,
    }


@pytest.fixture
def ensemble_config_with_aliases():
    return {
        "user": "test_user",
        "run_directory": str(parent_dir / "./"),
        "config_table": str(parent_dir / "test_config_alias.csv"),
        "aliases": {"banana": "setup:exp"},
        "share_repo": True,
    }


@pytest.fixture
def ensemble_config_with_nested_aliases():
    return {
        "user": "test_user",
        "run_directory": str(parent_dir / "./"),
        "config_table": str(parent_dir / "test_config_alias_nested.csv"),
        "aliases": {"banana": "setup:exp", "apple": "one:two:three"},
        "share_repo": True,
    }


@pytest.fixture
def ensemble_config_with_nested_alises_conflict():
    return {
        "user": "test_user",
        "run_directory": str(parent_dir / "./"),
        "config_table": str(parent_dir / "test_config_alias_nested_conflict.csv"),
        "aliases": {"banana": "setup:exp", "apple": "one:two:three"},
        "share_repo": True,
    }


def test_csv_table_vars_valid(config_data_one, ensemble_config):
    """
    MemberLevelVars + EnsembleLevelVars yield a valid CSVTableVars instance.
    """
    member_level = MemberLevelVars(config_data=config_data_one["member_level"])
    ensemble_level = EnsembleLevelVars(**ensemble_config)

    result = CSVTableVars(member_level=member_level, ensemble_level=ensemble_level)
    assert isinstance(result, CSVTableVars)


def test_csv_table_vars_invalid(config_data_one, ensemble_config):
    """
    Passing an obviously wrong type for `ensemble_level` should raise ValidationError.
    """
    member_level = MemberLevelVars(config_data=config_data_one["member_level"])
    with pytest.raises(ValidationError):
        CSVTableVars(member_level=member_level, ensemble_level=int)


def test_read_table_txt_one(config_data_one, ensemble_config_txt_one):
    expected = OmegaConf.create({
        "setup": {
            "runid": ["run-01", "run-02"],
            "exp": ["cmip6-piControl", "cmip6-piControl"],
        }
    })

    member_level = MemberLevelVars(config_data=config_data_one["member_level"])
    ensemble_level = EnsembleLevelVars(**ensemble_config_txt_one)

    table_vars = CSVTableVars(member_level=member_level, ensemble_level=ensemble_level)
    assert table_vars.read_table() == expected


@pytest.mark.filterwarnings("ignore:\\x1b\\[[0-9;]*m?Overlapping keys:UserWarning")
def test_read_table_with_aliases(ensemble_config_with_aliases):
    config_data = OmegaConf.create({
        "member_level": {
            "setup": {
                "runid": ["run-01", "run-02"],
            },
            "banana": ["cmip6-piControl", "cmip6-piControl"]
        }
    })

    expected = [
        {'setup': {'runid': 'run-01', 'exp': 'cmip6-piControl'}},
        {'setup': {'runid': 'run-02', 'exp': 'cmip6-piControl'}}
    ]

    cfg = OmegaConf.create({
        "member_level": config_data["member_level"],
        "ensemble_level": ensemble_config_with_aliases
    })

    ensemble_config, table = load_config(cfg, show_diffs=False)
    assert table == expected, "Resolved table does not match expected output"


@pytest.mark.filterwarnings("ignore:\\x1b\\[[0-9;]*m?Overlapping keys:UserWarning")
def test_read_table_with_nested_aliases(
    ensemble_config_with_nested_aliases
):
    config_data = OmegaConf.create({
        "member_level": {
            "setup": {
                "runid": ["run-01", "run-02"],
                "exp": ["cmip6-piControl", "cmip6-piControl"],
            },
            "apple": [1, 2],
        }
    })

    cfg = OmegaConf.create({
        "member_level": config_data["member_level"],
        "ensemble_level": ensemble_config_with_nested_aliases
    })

    expected = [
        {'setup': {'runid': 'run-01', 'exp': 'cmip6-piControl'}, 'one': {'two': {'three': 1}}},
        {'setup': {'runid': 'run-02', 'exp': 'cmip6-piControl'}, 'one': {'two': {'three': 2}}}
    ]

    ensemble_config, table = load_config(cfg, show_diffs=False)
    assert table == expected, "Resolved table does not match expected output"


@pytest.mark.filterwarnings("ignore:\\x1b\\[[0-9;]*m?Overlapping keys:UserWarning")
def test_read_table_nested_alias_conflict(
    ensemble_config_with_nested_alises_conflict
):
    """This tests the case where a key is treated as both a parameter and a value.
    In this case, one:two is assigned a value in the csv table, but one:two:three is an alias
    for a sub-value. Such a situation raises a conflict.
    """
    config_data = OmegaConf.create({
        "member_level": {
            "setup": {
                "runid": ["run-01", "run-02"],
            },
            "banana": ["cmip6-piControl", "cmip6-piControl"],
            "apple": [1, 2],
        }
    })

    cfg = OmegaConf.create({
        "member_level": config_data["member_level"],
        "ensemble_level": ensemble_config_with_nested_alises_conflict
    })

    with pytest.raises(ConfigTypeError):
        # This should raise an error due to the conflict in aliases
        # where 'one:two' is both a parameter and an alias.
        ensemble_config, table = load_config(cfg, show_diffs=False)


@pytest.mark.filterwarnings("ignore:\\x1b\\[[0-9;]*m?Overlapping keys:UserWarning")
def test_get_table(config_data_two, setup_config, ensemble_config_txt_one):

    member_level = MemberLevelVars(config_data=config_data_two["member_level"])
    ensemble_level = EnsembleLevelVars(**ensemble_config_txt_one)

    expected = [
        {
            "setup": {
                **setup_config,
                "runid": "run-01",
                "exp": "cmip6-piControl",
            }
        },
        {
            "setup": {
                **setup_config,
                "runid": "run-02",
                "exp": "cmip6-piControl",
            }
        },
    ]

    text_table_vars = CSVTableVars(
        member_level=member_level,
        ensemble_level=ensemble_level
    )

    assert text_table_vars.get_table(show_diffs=True) == expected



# tests/test_yaml_vars.py
import pytest
from omegaconf import OmegaConf, DictConfig

from imsi.tools.ensemble.table_utils.table_model import YAMLVars
from imsi.tools.ensemble.table_utils.data_model import MemberLevelVars

VER = "v5.1.7-imsi"


def _base_setup():
    return {
        "ver": VER,
        "repo": "git@gitlab.science.gc.ca:CanESM/CanESM5.git",
        "fetch_method": "clone",
        "machine": "ppp6",
        "model": "canesm51_p1",
    }


@pytest.fixture
def cfg_two_runids() -> DictConfig:
    """Two explicit run IDs (run1, run2)"""
    return OmegaConf.create(
        {"member_level": {"setup": {**_base_setup(), "runid": ["run1", "run2"]}}}
    )


@pytest.fixture
def cfg_auto_runid() -> DictConfig:
    """
    Single sentinel 'auto-gen-runid' plus a list field (model) of len 2 → 2
    members with auto‑generated unique run IDs expected.
    """
    return OmegaConf.create(
        {
            "member_level": {
                "setup": {
                    **_base_setup(),
                    "model": ["canesm51_p1", "canesm51_p1"],
                    "runid": "auto-gen-runid",
                }
            }
        }
    )


def test_yamlvars_broadcast_two_members(cfg_two_runids):
    member = MemberLevelVars(config_data=cfg_two_runids["member_level"])
    yaml_vars = YAMLVars(member_level=member)

    table = yaml_vars.get_table(show_diffs=False)
    table_plain = [OmegaConf.to_container(cfg, resolve=True) for cfg in table]

    expected = [
        {"setup": {**_base_setup(), "runid": "run1"}},
        {"setup": {**_base_setup(), "runid": "run2"}},
    ]
    assert table_plain == expected


def test_yamlvars_auto_generated_runids(cfg_auto_runid):
    member = MemberLevelVars(config_data=cfg_auto_runid["member_level"])

    # expand_config emits a warning for auto‑generated IDs
    with pytest.warns():
        yaml_vars = YAMLVars(member_level=member)
        table = yaml_vars.get_table(show_diffs=False)

    runids = [cfg.setup.runid for cfg in table]

    # Expect two unique IDs that start with 'auto-gen-runid-'
    assert len(runids) == 2
    assert len(set(runids)) == 2
    assert all(str(rid).startswith("auto-gen-runid") for rid in runids)



from imsi.tools.ensemble.table_utils.data_model import EnsembleLevelVars
from pydantic import ValidationError
import pytest
from pathlib import Path

parent_dir = Path(__file__).resolve().parent


# test various class instantiations with some purposefully incorrect inputs
good_example_vars = {
    "user": "test_user",
    "run_directory": str(parent_dir / "."),
    "config_table": str(parent_dir / "config.yaml"),
    "aliases": {"alias1": "banana"},
    "share_repo": True,
}

bad_example_vars_user = {
    "user": 1,
    "run_directory": "/path/to/run",
    "config_table": "/path/to/config_table",
    "aliases": {"alias1": "banana"},
    "share_repo": True,
}

bad_example_vars_run_directory = {
    "user": "test_user",
    "run_directory": "123",
    "config_table": "/path/to/config_table",
    "aliases": {"alias1": "banana"},
    "share_repo": True,
}

bad_example_vars_config_table = {
    "user": "test_user",
    "run_directory": "/path/to/run",
    "config_table": "123",
    "aliases": {"alias1": "banana"},
    "share_repo": True,
}

bad_example_vars_aliases = {
    "user": "test_user",
    "run_directory": "/path/to/run",
    "config_table": "/path/to/config_table",
    "aliases": "banana",
    "share_repo": True,
}

bad_example_vars_share_repo = {
    "user": "test_user",
    "run_directory": "/path/to/run",
    "config_table": "/path/to/config_table",
    "aliases": {"alias1": "banana"},
    "share_repo": True,
}


def test_good_ensemble_level_vars():
    EnsembleLevelVars(**good_example_vars)


@pytest.mark.parametrize(
    "ensemble_level_vars",
    [
        (bad_example_vars_user),
        (bad_example_vars_run_directory),
        (bad_example_vars_config_table),
        (bad_example_vars_aliases),
        (bad_example_vars_share_repo),
    ],
)
def test_bad_ensemble_level_vars(ensemble_level_vars):
    with pytest.raises(ValidationError):
        EnsembleLevelVars(**ensemble_level_vars)



import pandas as pd
from pathlib import Path

directory_paths = Path('canesm').rglob('*')

# split pahts
split_paths = [str(path).split('/') for path in directory_paths]

max_depth = max(len(path) for path in split_paths)
normalized_paths = [path + [None] * (max_depth - len(path)) for path in split_paths]

df = pd.DataFrame(normalized_paths, columns=[f'Level {i}' for i in range(max_depth)])
